import contextlib
import json
import operator
import pathlib
from pathlib import Path
from typing import Any, Dict

import dask
import yaml

import xorq as xo
import xorq.common.utils.logging_utils as lu
import xorq.vendor.ibis as ibis
import xorq.vendor.ibis.expr.types as ir
from xorq.common.utils.graph_utils import find_all_sources
from xorq.expr.udf import InputType
from xorq.ibis_yaml.common import SchemaRegistry, TranslationContext
from xorq.ibis_yaml.config import config
from xorq.ibis_yaml.sql import generate_sql_plans
from xorq.ibis_yaml.translate import (
    translate_from_yaml,
    translate_to_yaml,
)
from xorq.ibis_yaml.utils import freeze
from xorq.vendor.ibis.backends import Profile
from xorq.vendor.ibis.common.collections import FrozenOrderedDict


class CleanDictYAMLDumper(yaml.SafeDumper):
    def represent_frozenordereddict(self, data):
        return self.represent_dict(dict(data))

    def ignore_aliases(self, data):
        return True

    def represent_enum(self, data):
        return self.represent_scalar("tag:yaml.org,2002:str", data.name)

    def represent_ibis_schema(self, data):
        schema_dict = {name: str(dtype) for name, dtype in zip(data.names, data.types)}
        return self.represent_mapping("tag:yaml.org,2002:map", schema_dict)


CleanDictYAMLDumper.add_representer(
    FrozenOrderedDict, CleanDictYAMLDumper.represent_frozenordereddict
)

CleanDictYAMLDumper.add_representer(
    ibis.Schema, CleanDictYAMLDumper.represent_ibis_schema
)

CleanDictYAMLDumper.add_representer(InputType, CleanDictYAMLDumper.represent_enum)


class ArtifactStore:
    def __init__(self, root_path: pathlib.Path):
        self.root_path = (
            Path(root_path) if not isinstance(root_path, Path) else root_path
        )
        self.root_path.mkdir(parents=True, exist_ok=True)

    def get_path(self, *parts) -> pathlib.Path:
        return self.root_path.joinpath(*parts)

    def ensure_dir(self, *parts) -> pathlib.Path:
        path = self.get_path(*parts)
        path.mkdir(parents=True, exist_ok=True)
        return path

    def _read(self, read_f, *parts):
        path = self.get_path(*parts)
        with path.open("r") as f:
            return read_f(f)

    def read_yaml(self, *path_parts) -> Dict[str, Any]:
        return self._read(yaml.safe_load, *path_parts)

    def read_json(self, *path_parts) -> Dict[str, Any]:
        return self._read(json.load, *path_parts)

    def read_text(self, *path_parts) -> str:
        return self._read(operator.methodcaller("read"), *path_parts)

    @contextlib.contextmanager
    def _write(self, *path_parts):
        path = self.get_path(*path_parts)
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open("w") as f:
            yield (path, f)

    def write_yaml(self, data: Dict[str, Any], *path_parts) -> pathlib.Path:
        with self._write(*path_parts) as (path, f):
            yaml.dump(
                data,
                f,
                Dumper=CleanDictYAMLDumper,
                default_flow_style=False,
                sort_keys=False,
            )
        return path

    def write_text(self, content: str, *path_parts) -> pathlib.Path:
        with self._write(*path_parts) as (path, f):
            f.write(content)
        return path

    def exists(self, *path_parts) -> bool:
        return self.get_path(*path_parts).exists()

    def get_expr_hash(self, expr) -> str:
        expr_hash = dask.base.tokenize(expr)
        hash_length = config.hash_length
        return expr_hash[:hash_length]

    def save_yaml(self, yaml_dict: Dict[str, Any], expr_hash, filename) -> pathlib.Path:
        return self.write_yaml(yaml_dict, expr_hash, filename)

    def load_yaml(self, expr_hash: str, filename) -> Dict[str, Any]:
        return self.read_yaml(expr_hash, filename)

    def get_build_path(self, expr_hash: str) -> pathlib.Path:
        return self.ensure_dir(expr_hash)


class YamlExpressionTranslator:
    def __init__(self):
        pass

    def to_yaml(
        self,
        expr: ir.Expr,
        profiles=None,
    ) -> Dict[str, Any]:
        context = TranslationContext(
            schema_registry=SchemaRegistry(),
            profiles=freeze(profiles or {}),
        )

        schema_ref = context.schema_registry._register_expr_schema(expr)

        expr_dict = translate_to_yaml(expr, context)
        expr_dict = freeze({**dict(expr_dict), "schema_ref": schema_ref})

        context = context.finalize_definitions()

        return freeze(
            {
                "definitions": context.definitions,
                "expression": expr_dict,
            }
        )

    def from_yaml(
        self,
        yaml_dict: Dict[str, Any],
        profiles=None,
    ) -> ir.Expr:
        context = TranslationContext(
            schema_registry=SchemaRegistry(),
            profiles=freeze(profiles or {}),
        )

        context = context.update_definitions(freeze(yaml_dict.get("definitions", {})))

        expr_dict = freeze(yaml_dict["expression"])
        return translate_from_yaml(expr_dict, freeze(context))


class BuildManager:
    def __init__(self, build_dir: pathlib.Path):
        self.artifact_store = ArtifactStore(build_dir)
        self.profiles = {}

    def _write_sql_file(self, sql: str, expr_hash: str, query_name: str) -> str:
        hash_length = config.hash_length
        sql_hash = dask.base.tokenize(sql)[:hash_length]
        filename = f"{sql_hash}.sql"
        sql_path = self.artifact_store.get_build_path(expr_hash) / filename
        sql_path.write_text(sql)
        return filename

    def _process_sql_plans(
        self, sql_plans: Dict[str, Any], expr_hash: str
    ) -> Dict[str, Any]:
        updated_plans = {"queries": {}}

        for query_name, query_info in sql_plans["queries"].items():
            sql_filename = self._write_sql_file(
                query_info["sql"], expr_hash, query_name
            )

            updated_query_info = query_info.copy()
            updated_query_info["sql_file"] = sql_filename
            updated_query_info.pop("sql")
            updated_plans["queries"][query_name] = updated_query_info

        return updated_plans

    def _make_metadata(self) -> str:
        metadata = {
            "current_library_version": xo.__version__,
            "metadata_version": "0.0.0",  # TODO: make it a real thing
        }
        if lu._git_is_present():
            git_state = lu.get_git_state(hash_diffs=False)
            metadata["git_state"] = git_state

        metadata_json = json.dumps(metadata, indent=2)

        return metadata_json

    def _process_deferred_reads(
        self, deferred_reads: Dict[str, Any], expr_hash: str
    ) -> Dict[str, Any]:
        updated_reads = {"reads": {}}

        for read_name, read_info in deferred_reads["reads"].items():
            sql_filename = self._write_sql_file(read_info["sql"], expr_hash, read_name)

            updated_read_info = read_info.copy()
            updated_read_info["sql_file"] = sql_filename
            updated_read_info.pop("sql")
            updated_reads["reads"][read_name] = updated_read_info

        return updated_reads

    def compile_expr(self, expr: ir.Expr) -> str:
        expr_hash = self.artifact_store.get_expr_hash(expr)

        backends = find_all_sources(expr)
        profiles = {
            backend._profile.hash_name: {
                **backend._profile.as_dict(),
                "kwargs_tuple": dict(backend._profile.as_dict()["kwargs_tuple"]),
            }
            for backend in backends
        }

        translator = YamlExpressionTranslator()
        yaml_dict = translator.to_yaml(expr, profiles)
        self.artifact_store.save_yaml(yaml_dict, expr_hash, "expr.yaml")
        self.artifact_store.save_yaml(profiles, expr_hash, "profiles.yaml")

        sql_plans, deferred_reads = generate_sql_plans(expr)

        updated_sql_plans = self._process_sql_plans(sql_plans, expr_hash)
        self.artifact_store.save_yaml(updated_sql_plans, expr_hash, "sql.yaml")

        updated_deferred_reads = self._process_deferred_reads(deferred_reads, expr_hash)
        self.artifact_store.save_yaml(
            updated_deferred_reads, expr_hash, "deferred_reads.yaml"
        )

        metadata_json = self._make_metadata()
        self.artifact_store.write_text(metadata_json, expr_hash, "metadata.json")

        return expr_hash

    def load_expr(self, expr_hash: str) -> ir.Expr:
        profiles_dict = self.artifact_store.load_yaml(expr_hash, "profiles.yaml")

        def f(values):
            dct = dict(values)
            if isinstance(dct["kwargs_tuple"], dict):
                dct["kwargs_tuple"] = tuple(dct["kwargs_tuple"].items())
            else:
                dct["kwargs_tuple"] = tuple(map(tuple, dct["kwargs_tuple"]))
            return dct

        profiles = {
            profile: Profile(**f(values)).get_con()
            for profile, values in profiles_dict.items()
        }
        translator = YamlExpressionTranslator()

        yaml_dict = self.artifact_store.load_yaml(expr_hash, "expr.yaml")
        return translator.from_yaml(yaml_dict, profiles=profiles)

    # TODO: maybe change name
    def load_sql_plans(self, expr_hash: str) -> Dict[str, Any]:
        return self.artifact_store.load_yaml(expr_hash, "sql.yaml")

    def load_deferred_reads(self, expr_hash: str) -> Dict[str, Any]:
        return self.artifact_store.load_yaml(expr_hash, "deferred_reads.yaml")



from pathlib import Path

import pytest

import xorq as xo


@pytest.fixture(scope="session")
def csv_dir():
    root = Path(__file__).absolute().parents[5]
    data_dir = root / "ci" / "ibis-testing-data" / "csv"
    return data_dir


def test_register_read_csv(csv_dir):
    # this will use ls.options.backend: do we want to clear it out between function invocations?
    api_batting = xo.read_csv(csv_dir / "batting.csv", table_name="api_batting")
    result = xo.execute(api_batting)

    assert result is not None


def test_register_read_parquet(parquet_dir):
    # this will use ls.options.backend: do we want to clear it out between function invocations?
    api_batting = xo.read_parquet(
        parquet_dir / "batting.parquet", table_name="api_batting"
    )
    result = xo.execute(api_batting)

    assert result is not None


@pytest.mark.xfail(reason="No purpose with no registration api")
def test_executed_on_original_backend(parquet_dir, csv_dir, mocker):
    con = xo.config._backend_init()
    spy = mocker.spy(con, "execute")

    parquet_table = xo.read_parquet(parquet_dir / "batting.parquet")[
        lambda t: t.yearID == 2015
    ]

    csv_table = xo.read_csv(csv_dir / "batting.csv")[lambda t: t.yearID == 2014]

    expr = parquet_table.join(
        csv_table,
        "playerID",
    )

    assert xo.execute(expr) is not None
    assert spy.call_count == 1


def test_read_postgres():
    uri = "postgres://postgres:postgres@localhost:5432/ibis_testing"
    t = xo.read_postgres(uri, table_name="batting")
    res = xo.execute(t)

    assert res is not None and len(res)


@pytest.mark.xfail
def test_read_sqlite(tmp_path):
    import sqlite3

    xo.options.interactive = True
    db_path = tmp_path / "sqlite.db"
    with sqlite3.connect(db_path) as sq3:
        sq3.execute("DROP TABLE IF EXISTS t")
        sq3.execute("CREATE TABLE t (a INT, b TEXT)")
        sq3.execute("INSERT INTO t VALUES (1, 'a'), (2, 'b'), (3, 'c')")
    t = xo.read_sqlite(path=db_path, table_name="t")
    res = xo.execute(t)

    assert res is not None and len(res)


@pytest.mark.parametrize(
    ("with_repartition_file_scans", "keep_partition_by_columns"),
    [(True, True), (True, False), (False, True), (False, False)],
)
def test_with_config(
    with_repartition_file_scans, keep_partition_by_columns, parquet_dir
):
    import pandas as pd

    from xorq import SessionConfig

    session_config = (
        SessionConfig()
        .with_repartition_file_scans(with_repartition_file_scans)
        .set(
            "datafusion.execution.keep_partition_by_columns",
            str(keep_partition_by_columns).lower(),
        )
    )

    con = xo.connect(session_config=session_config)

    expr = con.read_parquet(parquet_dir / "batting.parquet").limit(10)
    result = expr.execute()

    assert isinstance(result, pd.DataFrame)
    assert len(result) == 10



from operator import methodcaller

import pandas as pd
import pyarrow as pa
import pytest

import xorq as xo
from xorq.caching import ParquetStorage, SourceStorage
from xorq.expr.relations import register_and_transform_remote_tables
from xorq.vendor import ibis
from xorq.vendor.ibis import _


expected_tables = (
    "array_types",
    "astronauts",
    "awards_players",
    "awards_players_special_types",
    "batting",
    "diamonds",
    "functional_alltypes",
    "geo",
    "geography_columns",
    "geometry_columns",
    "json_t",
    "map",
    "spatial_ref_sys",
    "topk",
    "tzone",
)


@pytest.fixture(scope="session")
def pg():
    conn = xo.postgres.connect_env()
    yield conn
    remove_unexpected_tables(conn)


@pytest.fixture(scope="session")
def trino_table():
    return (
        xo.trino.connect(database="tpch", schema="sf1")
        .table("orders")
        .cast({"orderdate": "date"})
    )


def make_merged(expr):
    agg = expr.group_by(["custkey", "orderdate"]).agg(
        _.totalprice.sum().name("totalprice")
    )
    w = ibis.window(group_by="custkey", order_by="orderdate")
    windowed = (
        agg.mutate(_.totalprice.cumsum().over(w).name("curev"))
        .mutate(_.curev.lag(1).over(w).name("curev@t-1"))
        .select(["custkey", "orderdate", "curev", "curev@t-1"])
    )
    merged = expr.asof_join(
        windowed,
        on="orderdate",
        predicates="custkey",
    ).select(
        [expr[c] for c in expr.columns]
        + [windowed[c] for c in windowed.columns if c not in expr.columns]
    )
    return merged


def remove_unexpected_tables(dirty):
    # drop tables
    for table in dirty.list_tables():
        if table not in expected_tables:
            dirty.drop_table(table, force=True)

    # drop view
    for table in dirty.list_tables():
        if table not in expected_tables:
            dirty.drop_view(table, force=True)

    if sorted(dirty.list_tables()) != sorted(expected_tables):
        raise ValueError


def test_multiple_record_batches(pg):
    con = xo.connect()

    table = pg.table("batting")
    left = con.register(table.to_pyarrow_batches(), "batting_0")
    right = con.register(table.to_pyarrow_batches(), "batting_1")

    expr = (
        left.join(right, "playerID")
        .limit(15)
        .select(player_id="playerID", year_id="yearID_right")
        .cache(SourceStorage(source=con))
    )

    res = expr.execute()
    assert isinstance(res, pd.DataFrame)
    assert 0 < len(res) <= 15


@pytest.mark.parametrize("method", [xo.to_pyarrow, xo.to_pyarrow_batches, xo.execute])
def test_into_backend_simple(pg, method):
    con = xo.connect()
    expr = pg.table("batting").into_backend(con, "ls_batting")
    res = method(expr)

    if isinstance(res, pa.RecordBatchReader):
        res = next(res)

    assert len(res) > 0


@pytest.mark.parametrize("method", ["to_pyarrow", "to_pyarrow_batches", "execute"])
def test_into_backend_complex(pg, method):
    con = xo.connect()

    t = pg.table("batting").into_backend(con, "ls_batting")

    expr = (
        t.join(t, "playerID")
        .limit(15)
        .select(player_id="playerID", year_id="yearID_right")
        .cache(SourceStorage(source=con))
    )

    assert xo.to_sql(expr).count("batting") == 2
    res = methodcaller(method)(expr)

    if isinstance(res, pa.RecordBatchReader):
        res = next(res)

    assert 0 < len(res) <= 15


def test_double_into_backend_batches(pg):
    con = xo.connect()
    ddb_con = xo.duckdb.connect()

    t = pg.table("batting").into_backend(con, "ls_batting")

    expr = (
        t.join(t, "playerID")
        .limit(15)
        .into_backend(ddb_con)
        .select(player_id="playerID", year_id="yearID_right")
        .cache(SourceStorage(source=con))
    )

    res = expr.to_pyarrow_batches()
    res = next(res)

    assert len(res) == 15


@pytest.mark.benchmark
def test_into_backend_cache(pg, tmp_path):
    con = xo.connect()
    ddb_con = xo.duckdb.connect()

    t = pg.table("batting").into_backend(con, "ls_batting")

    expr = (
        t.join(t, "playerID")
        .limit(15)
        .cache(SourceStorage(source=con))
        .into_backend(ddb_con)
        .select(player_id="playerID", year_id="yearID_right")
        .cache(ParquetStorage(source=ddb_con, path=tmp_path))
    )

    res = expr.execute()
    assert 0 < len(res) <= 15


def test_into_backend_duckdb(pg):
    ddb = xo.duckdb.connect()
    t = pg.table("batting").into_backend(ddb, "ls_batting")
    expr = (
        t.join(t, "playerID")
        .limit(15)
        .select(player_id="playerID", year_id="yearID_right")
    )

    expr, created = register_and_transform_remote_tables(expr)
    query = ibis.to_sql(expr, dialect="duckdb")

    res = ddb.con.sql(query).df()

    assert query.count("ls_batting") == 2
    assert 0 < len(res) <= 15
    assert len(created) == 3


def test_into_backend_duckdb_expr(pg):
    ddb = xo.duckdb.connect()
    t = pg.table("batting").into_backend(ddb, "ls_batting")
    expr = t.join(t, "playerID").limit(15).select(_.playerID * 2)

    expr, created = register_and_transform_remote_tables(expr)
    query = ibis.to_sql(expr, dialect="duckdb")

    res = ddb.con.sql(query).df()

    assert query.count("ls_batting") == 2
    assert 0 < len(res) <= 15
    assert len(created) == 3


def test_into_backend_duckdb_trino(trino_table):
    db_con = xo.duckdb.connect()
    expr = trino_table.head(10_000).into_backend(db_con).pipe(make_merged)

    expr, created = register_and_transform_remote_tables(expr)
    query = ibis.to_sql(expr, dialect="duckdb")

    df = db_con.con.sql(query).df()  # to bypass execute hotfix

    assert isinstance(df, pd.DataFrame)
    assert len(df) > 0
    assert len(created) == 3


def test_multiple_into_backend_duckdb_xorq(trino_table):
    db_con = xo.duckdb.connect()
    ls_con = xo.connect()

    expr = (
        trino_table.head(10_000)
        .into_backend(db_con)
        .pipe(make_merged)
        .into_backend(ls_con)[lambda t: t.orderstatus == "F"]
    )

    expr, created = register_and_transform_remote_tables(expr)

    df = expr.execute()

    assert isinstance(df, pd.DataFrame)
    assert len(df) > 0
    assert len(created) == 2


@pytest.mark.benchmark
def test_into_backend_duckdb_trino_cached(trino_table, tmp_path):
    db_con = xo.duckdb.connect()
    expr = (
        trino_table.head(10_000)
        .into_backend(db_con)
        .pipe(make_merged)
        .cache(ParquetStorage(path=tmp_path))
    )
    df = expr.execute()
    assert isinstance(df, pd.DataFrame)
    assert len(df) > 0


def test_into_backend_to_pyarrow_batches(trino_table):
    db_con = xo.duckdb.connect()
    df = (
        trino_table.head(10_000)
        .into_backend(db_con)
        .pipe(make_merged)
        .to_pyarrow_batches()
        .read_pandas()
    )
    assert not df.empty


def test_to_pyarrow_batches_simple(pg):
    con = xo.duckdb.connect()

    t = pg.table("batting").into_backend(con, "ls_batting")

    expr = (
        t.join(t, "playerID")
        .limit(15)
        .select(player_id="playerID", year_id="yearID_right")
    )

    df = expr.to_pyarrow_batches().read_pandas()
    assert not df.empty



import re

import pytest

import xorq as xo
from xorq import _
from xorq.caching import (
    ParquetSnapshotStorage,
    ParquetStorage,
    SourceSnapshotStorage,
    SourceStorage,
)
from xorq.common.utils.defer_utils import deferred_read_csv


def test_into_backend(batting):
    xo.options.interactive = False

    ddb_con = xo.duckdb.connect()

    t = batting.filter(batting.yearID == 2015).into_backend(ddb_con, "ls_batting")

    expr = (
        t.join(t, "playerID")
        .limit(15)
        .select(player_id="playerID", year_id="yearID_right")
    )

    assert "RemoteTable[r1, name=ls_batting]" in repr(expr)


@pytest.mark.parametrize(
    "storage, strategy, parquet",
    [
        pytest.param(ParquetStorage(), "modification_time", True, id="parquet_storage"),
        pytest.param(
            ParquetSnapshotStorage(), "snapshot", True, id="parquet_snapshot_storage"
        ),
        pytest.param(SourceStorage(), "modification_time", False, id="source_storage"),
        pytest.param(
            SourceSnapshotStorage(), "snapshot", False, id="source_snapshot_storage"
        ),
    ],
)
def test_cache(batting, storage, strategy, parquet):
    xo.options.interactive = False

    expr = (
        batting.join(batting, "playerID")
        .limit(15)
        .select(player_id="playerID", year_id="yearID_right")
        .cache(storage)
    )

    pattern = rf"CachedNode\[r\d+, strategy={strategy}, parquet={parquet}, source"

    assert re.search(pattern, repr(expr))


def test_read():
    xo.options.interactive = False

    csv_name = "iris"
    csv_path = xo.options.pins.get_path(csv_name)

    # we can work with a pandas expr without having read it yet
    pd_con = xo.pandas.connect()

    expr = deferred_read_csv(con=pd_con, path=csv_path, table_name=csv_name).filter(
        _.sepal_length > 6
    )

    pattern = r"Read\[name=iris, method_name=read_csv, source=pandas-\d+\]"

    assert re.search(pattern, repr(expr))



import pandas as pd
import pytest

import xorq as xo


def test_read_csv(con):
    name = "iris"
    table_name = f"testing-{name}"
    path = xo.options.pins.get_path(name)
    assert table_name not in con.tables
    t = con.read_csv(path, table_name)
    assert table_name in con.tables
    assert xo.execute(t).equals(pd.read_csv(path))


def test_read_csv_raises(con):
    name = "iris"
    table_name = f"testing-{name}"
    path = xo.options.pins.get_path(name)
    assert table_name not in con.tables
    with pytest.raises(
        ValueError, match="If `table_name` is not provided, `temporary` must be True"
    ):
        con.read_csv(path)
    assert table_name not in con.tables


def test_read_csv_temporary(con):
    name = "iris"
    table_name = f"testing-{name}"
    path = xo.options.pins.get_path(name)
    assert table_name not in con.tables
    t = con.read_csv(path, temporary=True)
    assert t.op().name in con.tables
    assert xo.execute(t).equals(pd.read_csv(path))


def test_read_csv_named_temporary(con):
    name = "iris"
    table_name = f"testing-{name}"
    path = xo.options.pins.get_path(name)
    assert table_name not in con.tables
    t = con.read_csv(path, table_name, temporary=True)
    assert table_name == t.op().name
    assert table_name in con.tables
    assert xo.execute(t).equals(pd.read_csv(path))


def test_read_parquet(con):
    name = "astronauts"
    table_name = f"testing-{name}"
    path = xo.options.pins.get_path(name)
    assert table_name not in con.tables
    t = con.read_parquet(path, table_name)
    assert table_name in con.tables
    assert xo.execute(t).equals(pd.read_parquet(path))


def test_read_parquet_raises(con):
    name = "astronauts"
    table_name = f"testing-{name}"
    path = xo.options.pins.get_path(name)
    assert table_name not in con.tables
    with pytest.raises(
        ValueError, match="If `table_name` is not provided, `temporary` must be True"
    ):
        con.read_parquet(path)
    assert table_name not in con.tables


def test_read_parquet_temporary(con):
    name = "astronauts"
    table_name = f"testing-{name}"
    path = xo.options.pins.get_path(name)
    assert table_name not in con.tables
    t = con.read_parquet(path, temporary=True)
    assert t.op().name in con.tables
    assert xo.execute(t).equals(pd.read_parquet(path))


def test_read_parquet_named_temporary(con):
    name = "astronauts"
    table_name = f"testing-{name}"
    path = xo.options.pins.get_path(name)
    assert table_name not in con.tables
    t = con.read_parquet(path, table_name, temporary=True)
    assert table_name == t.op().name
    assert table_name in con.tables
    assert xo.execute(t).equals(pd.read_parquet(path))



from __future__ import annotations

import gzip

import pandas as pd
import pyarrow as pa
import pyarrow.dataset as ds
import pytest

import xorq as xo
from xorq.tests.util import assert_frame_equal


@pytest.fixture
def gzip_csv(data_dir, tmp_path):
    basename = "diamonds.csv"
    f = tmp_path.joinpath(f"{basename}.gz")
    data = data_dir.joinpath("csv", basename).read_bytes()
    f.write_bytes(gzip.compress(data))
    return str(f.absolute())


def test_register_csv(con, data_dir):
    fname = "diamonds.csv"
    table_name = "diamonds"
    table = con.read_csv(data_dir / "csv" / fname, table_name=table_name)
    assert any(table_name in t for t in con.list_tables())
    assert table.count().execute() > 0


@pytest.mark.xfail
def test_register_csv_gz(con, data_dir, gzip_csv):
    # TODO review why there is a regression
    table = con.read_csv(gzip_csv, table_name="diamonds")
    assert table.count().execute() > 0


def test_register_with_dotted_name(con, data_dir, tmp_path):
    basename = "foo.bar.baz/diamonds.csv"
    f = tmp_path.joinpath(basename)
    f.parent.mkdir()
    data = data_dir.joinpath("csv", "diamonds.csv").read_bytes()
    f.write_bytes(data)
    table = con.read_csv(str(f.absolute()), table_name="diamonds")
    assert table.count().execute() > 0


def test_register_parquet(con, data_dir):
    fname = "functional_alltypes.parquet"
    table_name = "funk_all"
    table = con.read_parquet(data_dir / "parquet" / fname, table_name=table_name)

    assert any(table_name in t for t in con.list_tables())
    assert table.count().execute() > 0


def test_read_parquet_with_custom_file_extension(con, tmp_path):
    parquet_path = (tmp_path / "data").with_suffix(".arquet")

    expected = pd.DataFrame([{"maxitem": 43182839, "n": 1000}])
    expected.to_parquet(parquet_path)

    actual = xo.read_parquet(parquet_path).execute()

    assert_frame_equal(expected, actual)


def test_read_csv_with_custom_file_extension(con, tmp_path):
    csv_path = (tmp_path / "data").with_suffix(".acsv")

    expected = pd.DataFrame([{"maxitem": 43182839, "n": 1000}])
    expected.to_csv(csv_path, index=False)

    actual = xo.read_csv(csv_path).execute()

    assert_frame_equal(expected, actual)


def test_read_csv(con, data_dir):
    t = con.read_csv(data_dir / "csv" / "functional_alltypes.csv")
    assert t.count().execute()


def test_read_csv_from_url(con):
    t = con.read_csv(
        "https://raw.githubusercontent.com/ibis-project/testing-data/refs/heads/master/csv/astronauts.csv"
    )
    assert t.head().execute() is not None


@pytest.mark.s3
def test_read_csv_from_s3(con):
    t = con.read_csv(
        "s3://opendata-downloads/opa_properties_public.csv",
    )
    assert t.head().execute() is not None


def test_read_parquet(con, data_dir):
    t = con.read_parquet(data_dir / "parquet" / "functional_alltypes.parquet")
    assert t.count().execute()


def test_read_parquet_from_url(con):
    t = con.read_parquet(
        "https://nasa-avionics-data-ml.s3.us-east-2.amazonaws.com/Tail_652_1_parquet/652200101120916.16p0.parquet",
        table_name="nasa_table",
    )

    assert "nasa_table" in con.list_tables()
    assert t.op().name == "nasa_table"
    assert t.head().execute() is not None


def test_register_table(con):
    tab = pa.table({"x": [1, 2, 3]})
    con.create_table("my_table", tab)
    assert con.table("my_table").x.sum().execute() == 6


def test_register_pandas(con):
    df = pd.DataFrame({"x": [1, 2, 3]})
    con.create_table("my_pandas_table", df)
    assert con.table("my_pandas_table").x.sum().execute() == 6


def test_register_batches(con):
    batch = pa.record_batch([pa.array([1, 2, 3])], names=["x"])
    con.create_table("my_batch_table", batch)
    assert con.table("my_batch_table").x.sum().execute() == 6


def test_register_dataset(con):
    tab = pa.table({"x": [1, 2, 3]})
    dataset = ds.InMemoryDataset(tab)
    t = xo.memtable(dataset, name="my_table")
    assert con.execute(t.x.sum()) == 6


def test_register_memtable(con):
    data = pd.DataFrame({"a": [1, 2, 3, 4, 5], "b": [2, 3, 4, 5, 6]})
    t = xo.memtable(data)
    assert con.execute(t.a.sum()) == 15



from typing import Any, Dict, List, Tuple, TypedDict

import xorq.vendor.ibis as ibis
import xorq.vendor.ibis.expr.operations as ops
import xorq.vendor.ibis.expr.types as ir
from xorq.common.utils.graph_utils import find_all_sources, walk_nodes
from xorq.expr.relations import Read, RemoteTable


class QueryInfo(TypedDict):
    engine: str
    profile_name: str
    sql: str


class SQLPlans(TypedDict):
    queries: Dict[str, QueryInfo]


class DeferredReadsPlan(TypedDict):
    reads: Dict[str, QueryInfo]


def find_relations(expr: ir.Expr) -> List[str]:
    def get_name(node):
        name = None
        if isinstance(node, RemoteTable):
            name = node.name
        elif isinstance(node, Read):
            name = node.make_unbound_dt().name
        elif isinstance(node, ops.DatabaseTable):
            name = node.name
        return name

    node_types = (RemoteTable, Read, ops.DatabaseTable)
    nodes = walk_nodes(node_types, expr)
    relations = list(set(filter(None, map(get_name, nodes))))
    return relations


def find_tables(expr: ir.Expr) -> Tuple[Dict[str, QueryInfo], Dict[str, QueryInfo]]:
    remote_tables: Dict[str, QueryInfo] = {}
    deferred_reads: Dict[str, QueryInfo] = {}

    node_types = (RemoteTable, Read)
    nodes = walk_nodes(node_types, expr)

    for node in nodes:
        if isinstance(node, RemoteTable):
            remote_expr = node.remote_expr
            backends = find_all_sources(node)
            if len(backends) > 1:
                backends = tuple(
                    x for x in backends if x != node.to_expr()._find_backend()
                )
            for backend in backends:
                engine_name = backend.name
                profile_name = backend._profile.hash_name
                key = f"{node.name}"
                remote_tables[key] = {
                    "engine": engine_name,
                    "profile_name": profile_name,
                    "relations": find_relations(remote_expr),
                    "sql": ibis.to_sql(remote_expr).strip(),
                    "options": {},
                }
        elif isinstance(node, Read):
            backend = node.source
            if backend is not None:
                dt = node.make_unbound_dt()
                key = dt.name
                deferred_reads[key] = {
                    "engine": backend.name,
                    "profile_name": backend._profile.hash_name,
                    "relations": [dt.name],
                    "sql": ibis.to_sql(dt.to_expr()).strip(),
                    "options": get_read_options(node),
                }
    return remote_tables, deferred_reads


def get_read_options(read_instance) -> Dict[str, Any]:
    read_kwargs_list = [{k: v} for k, v in read_instance.read_kwargs]
    return {
        "method_name": read_instance.method_name,
        "name": read_instance.name,
        "read_kwargs": read_kwargs_list,
    }


def generate_sql_plans(expr: ir.Expr) -> Tuple[SQLPlans, DeferredReadsPlan]:
    remote_tables, deferred_reads = find_tables(expr)
    main_sql = ibis.to_sql(expr)
    backend = expr._find_backend()

    queries: Dict[str, QueryInfo] = {
        "main": {
            "engine": backend.name,
            "profile_name": backend._profile.hash_name,
            "relations": find_relations(expr),
            "sql": main_sql.strip(),
            "options": {},
        }
    }

    for table_name, info in remote_tables.items():
        queries[table_name] = info

    sql_plans: SQLPlans = {"queries": queries}
    deferred_reads_plans: DeferredReadsPlan = {"reads": deferred_reads}
    return sql_plans, deferred_reads_plans



import os

from xorq.vendor.ibis.backends.profiles import Profile, Profiles


print("=== XORQ PROFILES DEMONSTRATION ===\n")

# Set environment variables for database connection
print("Setting up environment variables...")
# This is a PostgreSQL database running at examples.letsql.com
os.environ["POSTGRES_DATABASE"] = "letsql"
os.environ["POSTGRES_HOST"] = "examples.letsql.com"
os.environ["POSTGRES_USER"] = "letsql"
os.environ["POSTGRES_PASSWORD"] = "letsql"
os.environ["POSTGRES_PORT"] = "5432"

# 1. Create a profile with environment variable references
print("\n1. Creating profile with environment variable references...")
profile = Profile(
    con_name="postgres",
    kwargs_tuple=(
        ("host", "${POSTGRES_HOST}"),
        ("port", 5432),
        ("database", "postgres"),
        ("user", "${POSTGRES_USER}"),
        ("password", "${POSTGRES_PASSWORD}"),
    ),
)

print(f"Profile representation:\n{profile}")

# 2. Save the profile with an alias
print("\n2. Saving profile with alias...")
path = profile.save(alias="postgres_example", clobber=True)
print(f"Profile saved to: {path}")

# 3. Load the profile
print("\n3. Loading profile from disk...")
loaded_profile = Profile.load("postgres_example")
print(f"Loaded profile representation:\n{loaded_profile}")

# 4. Create a connection from the profile
print("\n4. Creating connection from profile...")
connection = loaded_profile.get_con()
print("Connection successful!")

# 5. Verify connection works
tables = connection.list_tables()
print(f"Found tables: {tables[:5]}")

# 6. Verify connection's profile still has environment variables
print("\n5. Examining connection's profile...")
conn_profile = connection._profile
print(f"Connection profile representation:\n{conn_profile}")

# 7. Create a profile from existing connection
print("\n6. Creating new profile from connection...")
from_conn_profile = Profile.from_con(connection)
print(f"Profile from connection representation:\n{from_conn_profile}")

# 8. Save profile from connection with new alias
print("\n7. Saving profile from connection...")
from_conn_profile.save(alias="postgres_from_conn", clobber=True)

# 9. Working with multiple profiles
print("\n8. Working with multiple profiles...")
profiles = Profiles()
all_profiles = profiles.list()
print(f"Available profiles: {all_profiles}")

# 10. Clone a profile with modifications
print("\n9. Cloning profile with modifications...")
cloned_profile = profile.clone(**{"connect_timeout": 10})
print(f"Original profile representation:\n{profile}")
print(f"Cloned profile representation:\n{cloned_profile}")

# 11. Save the cloned profile
cloned_profile.save(alias="postgres_other_db", clobber=True)

# 12. Security verification
print("\n10. Security verification...")
print("Throughout this entire process, actual values of environment")
print("variables were never stored in profiles or exposed in output.")

# 13. Demonstrating how to explore with profiles
print("\n11. Exploring available profiles...")
profiles = Profiles()
for name in profiles.list():
    p = profiles.get(name)
    print(f"Profile: {name}")
    print(f"  - Connection: {p.get_con()}")
    print(f"  - Profile: {p}")



from pathlib import Path

import pandas as pd

import xorq as xo
from xorq.expr.relations import into_backend
from xorq.ml import make_quickgrove_udf, rewrite_quickgrove_expr


model_path = Path(xo.options.pins.get_path("diamonds-model"))

# source backend
pg = xo.postgres.connect_examples()
# local backend
con = xo.connect()
# load xgboost json model
# cannot have hyphen in the model name see #498
model = make_quickgrove_udf(model_path, model_name="diamonds_model")

t = (
    into_backend(pg.tables["diamonds"], con, "diamonds")
    .mutate(pred=model.on_expr)
    .filter(xo._.carat < 1)
    .select(xo._.pred)
)

t_pruned = rewrite_quickgrove_expr(t)

original = xo.execute(t)
pruned = xo.execute(t_pruned)

pd.testing.assert_frame_equal(original, pruned, rtol=3)



import operator
import pickle

import pandas as pd
import pyarrow as pa
import pytest
import toolz
import xgboost as xgb

import xorq as xo
import xorq.expr.datatypes as dt
from xorq import make_pandas_expr_udf, make_pandas_udf, udf
from xorq.tests.util import assert_frame_equal
from xorq.vendor.ibis import _
from xorq.vendor.ibis import literal as L
from xorq.vendor.ibis.selectors import of_type


pc = pytest.importorskip("pyarrow.compute")


@udf.agg.pyarrow
def my_mean(arr: dt.float64) -> dt.float64:
    return pc.mean(arr)


@udf.agg.pyarrow
def add_mean(a: dt.float64, b: dt.float64) -> dt.float64:
    return pc.mean(pc.add(a, b))


return_type = pa.struct(
    [
        ("a", pa.float64()),
        ("b", pa.float64()),
        ("c", pa.float64()),
    ]
)


@udf.agg.pyarrow
def centroid(a: dt.float64, b: dt.float64, c: dt.float64) -> return_type:
    return pa.scalar((pc.mean(a), pc.mean(b), pc.mean(c)), type=return_type)


@udf.agg.pyarrow
def centroid_list(a: dt.float64, b: dt.float64, c: dt.float64) -> pa.list_(
    pa.float64()
):
    return pa.scalar([pc.mean(a), pc.mean(b), pc.mean(c)], type=pa.list_(pa.float64()))


def test_udf_agg_pyarrow(ls_con, batting):
    batting = ls_con.register(xo.execute(batting), "pg-batting")
    result = my_mean(batting.G).execute()

    assert result == xo.execute(batting.G).mean()


def test_multiple_arguments_udf_agg_pyarrow(ls_con, batting):
    batting = ls_con.register(xo.execute(batting), "pg-batting")
    actual = add_mean(batting.G, batting.G).execute()
    expected = xo.execute(batting.G)
    expected = (expected + expected).mean()

    assert actual == expected


def test_multiple_arguments_struct_udf_agg_pyarrow(ls_con, batting):
    from math import isclose

    batting = ls_con.register(xo.execute(batting), "pg-batting")
    actual = centroid(batting.G, batting.G, batting.G).execute()
    expected = xo.execute(batting.G).mean()

    assert all(isclose(value, expected) for value in actual.values())


def test_multiple_arguments_list_udf_agg_pyarrow(ls_con, batting):
    from math import isclose

    batting = ls_con.register(xo.execute(batting), "pg-batting")
    actual = centroid_list(batting.G, batting.G, batting.G).execute()
    expected = xo.execute(batting.G).mean()

    assert all(isclose(value, expected) for value in actual)


@udf.agg.pyarrow
def my_sum(arr: dt.float64) -> dt.float64:
    return pc.sum(arr)


def test_group_by_udf_agg_pyarrow(ls_con, alltypes_df):
    alltypes = ls_con.register(alltypes_df, "pg-alltypes")

    expr = (
        alltypes[_.string_col == "1"]
        .mutate(x=L(1, "int64"))
        .group_by(_.x)
        .aggregate(sum=my_sum(_.double_col))
    )

    result = xo.execute(expr).astype({"x": "int64"})
    expected = (
        alltypes_df.loc[alltypes_df.string_col == "1", :]
        .assign(x=1)
        .groupby("x")
        .double_col.sum()
        .rename("sum")
        .reset_index()
    )

    assert_frame_equal(result, expected, check_like=True)


def test_udf_agg_pandas_df(ls_con, alltypes):
    name = "sum_sum"
    alltypes = ls_con.register(xo.execute(alltypes), "pg-alltypes")
    cols = (by, _) = ["year", "month"]
    expr = alltypes

    @udf.agg.pandas_df(
        schema=expr[cols].schema(),
        return_type=dt.int64(),
        name=name,
    )
    def sum_sum(df):
        return df.sum().sum()

    actual = (
        expr.group_by(by)
        .agg(sum_sum(*(expr[c] for c in cols)).name(name))
        .execute()
        .sort_values(by, ignore_index=True)
    )
    actual2 = (
        expr.group_by(by)
        .agg(sum_sum.on_expr(expr).name(name))
        .execute()
        .sort_values(by, ignore_index=True)
    )
    expected = (
        alltypes[cols]
        .execute()
        .groupby(by)
        .apply(sum_sum.fn)
        .rename(name)
        .reset_index()
        .sort_values(by, ignore_index=True)
    )
    assert actual.equals(expected)
    assert actual2.equals(expected)


def test_udf_pandas_df(ls_con, batting):
    typ = "int64"
    name = "my_least"
    batting = ls_con.register(xo.execute(batting), "pg-batting")
    my_least = operator.methodcaller("min", axis=1)
    schema = batting.select(of_type(typ)).schema()
    return_type = dt.dtype(typ)
    udf = make_pandas_udf(my_least, schema, return_type, name=name)
    order_by = ("playerID", "yearID", "stint")

    from_builtin = ls_con.execute(
        batting.mutate(xo.least(*(xo._[name] for name in schema)).name(name)).order_by(
            order_by
        )
    )
    from_udf = ls_con.execute(
        batting.mutate(udf.on_expr(batting).name(name)).order_by(order_by)
    )
    from_pd = (
        ls_con.execute(batting)
        .assign(**{name: lambda t: my_least(t[list(schema)])})
        .sort_values(list(order_by), ignore_index=True)
    )
    assert from_builtin.equals(from_udf)
    assert from_builtin.equals(from_pd)


def test_pandas_expr_udf():
    @toolz.curry
    def train_xgboost_model(df, features, target, seed=0):
        param = {"max_depth": 4, "eta": 1, "objective": "binary:logistic", "seed": seed}
        num_round = 10
        X = df[list(features)]
        y = df[target]
        dtrain = xgb.DMatrix(X, y)
        bst = xgb.train(param, dtrain, num_boost_round=num_round)
        return bst

    def predict_xgboost_model(model, df):
        return model.predict(xgb.DMatrix(df))

    features = (
        "emp_length",
        "dti",
        "annual_inc",
        "loan_amnt",
        "fico_range_high",
        "cr_age_days",
    )
    target = "event_occurred"
    train_fn = train_xgboost_model(features=features, target=target)
    name = "predicted"
    typ = "float64"

    con = xo.connect()
    t = con.read_parquet(xo.config.options.pins.get_path("lending-club"))

    # manual run
    df = xo.execute(t)
    model = train_fn(df)
    from_pd = df.assign(
        **{name: predict_xgboost_model(model, df[list(features)])}
    ).astype({name: typ})

    # using expr-scalar-udf
    model_udaf = udf.agg.pandas_df(
        fn=toolz.compose(pickle.dumps, train_fn),
        schema=t[features + (target,)].schema(),
        return_type=dt.binary,
        name="xgboost_model",
    )
    computed_kwargs_expr = model_udaf.on_expr(t)
    predict_expr_udf = make_pandas_expr_udf(
        computed_kwargs_expr,
        fn=predict_xgboost_model,
        schema=t[features].schema(),
        return_type=dt.dtype(typ),
        name=name,
    )
    expr = t.mutate(predict_expr_udf.on_expr(t).name(name))
    from_ls = con.execute(expr)

    pd._testing.assert_frame_equal(from_ls, from_pd)



from typing import MutableMapping

import pandas as pd

import xorq.common.exceptions as com
from xorq.vendor.ibis.backends.pandas import Backend as IbisPandasBackend


class Backend(IbisPandasBackend):
    def do_connect(
        self,
        dictionary: MutableMapping[str, pd.DataFrame] | None = None,
    ) -> None:
        """Construct a client from a dictionary of pandas DataFrames.

        Parameters
        ----------
        dictionary
            An optional mapping of string table names to pandas DataFrames.

        Examples
        --------
        >>> import xorq as ls
        >>> ls.pandas.connect({"t": pd.DataFrame({"a": [1, 2, 3]})})  # doctest: +ELLIPSIS
        <ibis.backends.pandas.Backend object at 0x...>
        """
        self.dictionary = dictionary or {}
        self.schemas = {}

    def drop_table(self, name: str, *, force: bool = False) -> None:
        if not force and name in self.dictionary:
            raise com.XorqError(
                "Cannot drop existing table. Call drop_table with force=True to drop existing table."
            )
        del self.dictionary[name]
        del self.schemas[name]



from contextlib import contextmanager

import pytest
import sqlglot as sg
import sqlglot.expressions as sge

import xorq as xo
from xorq.vendor.ibis.util import gen_name


SU = pytest.importorskip("xorq.common.utils.snowflake_utils")


@pytest.fixture(scope="session")
def sf_con():
    return xo.snowflake.connect(
        # a database/schema we can trust exists
        database="SNOWFLAKE_SAMPLE_DATA",
        schema="TPCH_SF1",
        **SU.make_credential_defaults(),
        **SU.make_connection_defaults(),
        create_object_udfs=False,
    )


@pytest.fixture
def temp_catalog(sf_con):
    cat = gen_name("tmp_catalog")

    sf_con.create_catalog(cat)
    assert cat in sf_con.list_catalogs()

    yield cat

    sf_con.drop_catalog(cat)
    assert cat not in sf_con.list_catalogs()


@pytest.fixture
def temp_db(sf_con, temp_catalog):
    database = gen_name("tmp_database")

    sf_con.create_database(database, catalog=temp_catalog)
    assert database in sf_con.list_databases(catalog=temp_catalog)

    yield database

    sf_con.drop_database(database, catalog=temp_catalog)
    assert database not in sf_con.list_databases(catalog=temp_catalog)


@contextmanager
def inside_temp_schema(con, temp_catalog, temp_db):
    (prev_catalog, prev_db) = (con.current_catalog, con.current_database)
    con.raw_sql(
        sge.Use(
            kind="SCHEMA", this=sg.table(temp_db, db=temp_catalog, quoted=True)
        ).sql(dialect=con.name),
    )
    try:
        yield
    finally:
        con.raw_sql(
            sge.Use(
                kind="SCHEMA", this=sg.table(prev_db, db=prev_catalog, quoted=True)
            ).sql(dialect=con.name),
        )



from pathlib import Path
from typing import Any, Iterable, Mapping

import pandas as pd
import pyarrow as pa

from xorq.common.utils.rbr_utils import (
    make_filtered_reader,
)
from xorq.flight.action import (
    DropTableAction,
    DropViewAction,
    GetSchemaQueryAction,
    ListTablesAction,
    ReadParquetAction,
    TableInfoAction,
    VersionAction,
)
from xorq.flight.client import FlightClient
from xorq.vendor import ibis
from xorq.vendor.ibis import util
from xorq.vendor.ibis.backends.sql import SQLBackend
from xorq.vendor.ibis.expr import schema as sch
from xorq.vendor.ibis.expr import types as ir
from xorq.vendor.ibis.util import gen_name


class Backend(SQLBackend):
    @property
    def name(self):
        return "xorq_flight"

    @property
    def version(self) -> str:
        return self.con.do_action(VersionAction.name, options=self.con._options)[0]

    def create_table(
        self,
        name: str,
        obj: pd.DataFrame | pa.Table | ir.Table | None = None,
        *,
        schema: ibis.Schema | None = None,
        database: str | None = None,
        temp: bool = False,
        overwrite: bool = False,
    ) -> ir.Table:
        if isinstance(obj, pd.DataFrame):
            obj = pa.Table.from_pandas(obj)

        if isinstance(obj, pa.Table):
            obj = obj.to_batches()

        if isinstance(obj, ir.Table):
            obj = obj.to_pyarrow_batches()

    def _get_schema_using_query(self, query: str) -> sch.Schema:
        return self.con.do_action_one(
            GetSchemaQueryAction.name, action_body=query, options=self.con._options
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.con = None

    def do_connect(
        self,
        host="localhost",
        port=8815,
        username=None,
        password=None,
        tls_roots=None,
    ) -> None:
        self.con = FlightClient(
            host=host,
            port=port,
            username=username,
            password=password,
            tls_roots=tls_roots,
        )

    def get_schema(
        self,
        table_name: str,
        *,
        catalog: str | None = None,
        database: str | None = None,
    ) -> sch.Schema:
        args = {
            "table_name": table_name,
            "catalog": catalog,
            "database": database,
        }
        return self.con.do_action_one(
            TableInfoAction.name, action_body=args, options=self.con._options
        )

    def read_in_memory(
        self,
        source: pd.DataFrame | pa.Table | pa.RecordBatchReader,
        table_name: str | None = None,
    ) -> ir.Table:
        table_name = table_name or util.gen_name("read_in_memory")

        if isinstance(source, pa.Table):
            self.con.upload_data(table_name, source)
        elif isinstance(source, pa.RecordBatchReader):
            self.con.upload_batches(table_name, source)
        return self.table(table_name)

    def read_parquet(
        self,
        source_list: str | Iterable[str],
        table_name: str | None = None,
        **kwargs: Any,
    ) -> ir.Table:
        args = {
            "source_list": source_list,
            "table_name": table_name,
        }
        table_name = self.con.do_action_one(
            ReadParquetAction.name, action_body=args, options=self.con._options
        )
        return self.table(table_name)

    def register(
        self,
        source: str | Path | Any,
        table_name: str | None = None,
        **kwargs: Any,
    ) -> ir.Table:
        if isinstance(source, pd.DataFrame):
            source = pa.Table.from_pandas(source)
            source = pa.RecordBatchReader.from_batches(
                source.schema, source.to_batches()
            )

        return self.read_record_batches(source, table_name=table_name)

    def read_record_batches(self, source, table_name=None):
        table_name = table_name or gen_name("read_record_batches")
        self.con.upload_batches(table_name, source)
        return self.table(table_name)

    def list_tables(
        self, like: str | None = None, database: tuple[str, str] | str | None = None
    ) -> list[str]:
        kwargs = {
            "like": like,
            "database": database,
        }

        return self.con.do_action_one(
            ListTablesAction.name, action_body=kwargs, options=self.con._options
        )

    def drop_table(
        self,
        name: str,
        database: tuple[str, str] | str | None = None,
        force: bool = False,
    ) -> None:
        kwargs = {"name": name, "database": database, "force": force}
        self.con.do_action(
            DropTableAction.name, action_body=kwargs, options=self.con._options
        )

    def drop_view(
        self,
        name: str,
        *,
        database: str | None = None,
        schema: str | None = None,
        force: bool = False,
    ) -> None:
        kwargs = {"name": name, "database": database, "schema": schema, "force": force}
        self.con.do_action(
            DropViewAction.name, action_body=kwargs, options=self.con._options
        )

    def to_pyarrow_batches(
        self,
        expr: ir.Expr,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        chunk_size: int = 10_000,
        **_: Any,
    ) -> pa.ipc.RecordBatchReader:
        batches = self.con.execute_batches(
            expr, params=params, limit=limit, chunk_size=chunk_size
        )
        batches = make_filtered_reader(batches)
        return batches



import functools
import json
from pathlib import Path

import pandas as pd
import requests
import toolz

import xorq as xo


base_api_url = "https://hacker-news.firebaseio.com/v0"


@toolz.curry
def simple_disk_cache(f, cache_dir, serde):
    cache_dir = Path(cache_dir).absolute()
    cache_dir.mkdir(parents=True, exist_ok=True)

    def wrapped(**kwargs):
        name = ",".join(f"{key}={value}" for key, value in kwargs.items())
        path = cache_dir.joinpath(name)
        if path.exists():
            value = serde.loads(path.read_text())
        else:
            value = f(**kwargs)
            path.write_text(serde.dumps(value))
        return value

    return wrapped


def get_json(url):
    resp = requests.get(url)
    resp.raise_for_status()
    return resp.json()


@simple_disk_cache(cache_dir=Path("./hackernews-items"), serde=json)
def get_hackernews_item(*, item_id):
    return get_json(f"{base_api_url}/item/{item_id}.json")


@functools.cache
def get_hackernews_maxitem():
    return get_json(f"{base_api_url}/maxitem.json")


def get_hackernews_stories(maxitem, n):
    gen = (
        toolz.excepts(requests.exceptions.SSLError, get_hackernews_item)(
            item_id=item_id
        )
        for item_id in range(maxitem - n, maxitem)
    )
    gen = filter(None, gen)
    df = pd.DataFrame(gen).reindex(columns=schema_out)
    return df


@toolz.curry
def get_hackernews_stories_batch(df, filter=slice(None)):
    series = df.apply(lambda row: get_hackernews_stories(**row), axis=1)
    return (
        pd.concat(series.values, ignore_index=True).loc[filter].reset_index(drop=True)
    )


schema_in = xo.schema({"maxitem": int, "n": int})
schema_out = xo.schema(
    {
        "by": "string",
        "id": "int64",
        "parent": "float64",
        "text": "string",
        "time": "int64",
        "type": "string",
        "kids": "array<int64>",
        # "deleted": "bool",
        "descendants": "float64",
        "score": "float64",
        "title": "string",
        "url": "string",
        # "dead": "bool",
    }
)


do_hackernews_fetcher_udxf = xo.expr.relations.flight_udxf(
    process_df=get_hackernews_stories_batch(
        filter=lambda t: t.type.eq("story") & t.title.notnull()
    ),
    # process_df=get_hackernews_stories_batch,
    maybe_schema_in=schema_in.to_pyarrow(),
    maybe_schema_out=schema_out.to_pyarrow(),
    name="HackerNewsFetcher",
)



import base64
from typing import Any, Dict

import cloudpickle

import xorq.expr.datatypes as dt
import xorq.vendor.ibis.expr.operations as ops
import xorq.vendor.ibis.expr.rules as rlz
from xorq.expr.relations import FlightExpr, FlightUDXF
from xorq.ibis_yaml.common import (
    _translate_type,
    register_from_yaml_handler,
    translate_from_yaml,
    translate_to_yaml,
)
from xorq.ibis_yaml.utils import freeze
from xorq.vendor.ibis.common.annotations import Argument


def serialize_udf_function(fn: callable) -> str:
    pickled = cloudpickle.dumps(fn)
    encoded = base64.b64encode(pickled).decode("ascii")
    return encoded


def deserialize_udf_function(encoded_fn: str) -> callable:
    pickled = base64.b64decode(encoded_fn)
    return cloudpickle.loads(pickled)


@translate_to_yaml.register(ops.ScalarUDF)
def _scalar_udf_to_yaml(op: ops.ScalarUDF, compiler: Any) -> dict:
    input_type = getattr(op.__class__, "__input_type__", None)

    if input_type not in [ops.udf.InputType.BUILTIN, ops.udf.InputType.PYARROW]:
        raise NotImplementedError(
            f"Translation of UDFs with input type {getattr(op.__class__, '__input_type__', None)} is not supported"
        )
    arg_names = [
        name
        for name in dir(op)
        if not name.startswith("__") and name not in op.__class__.__slots__
    ]

    return freeze(
        {
            "op": "ScalarUDF",
            "func_name": op.__func_name__,
            "input_type": input_type,
            "args": [translate_to_yaml(arg, compiler) for arg in op.args],
            "type": _translate_type(op.dtype),
            "pickle": serialize_udf_function(op.__func__),
            "module": op.__module__,
            "class_name": op.__class__.__name__,
            "arg_names": arg_names,
        }
    )


@register_from_yaml_handler("ScalarUDF")
def _scalar_udf_from_yaml(yaml_dict: dict, compiler: any) -> any:
    # TODO: use make_pandas_udf (expr/udf.py)
    encoded_fn = yaml_dict.get("pickle")
    if not encoded_fn:
        raise ValueError("Missing pickle data for ScalarUDF")
    fn = deserialize_udf_function(encoded_fn)

    args = tuple(
        translate_from_yaml(arg, compiler) for arg in yaml_dict.get("args", [])
    )
    if not args:
        raise ValueError("ScalarUDF requires at least one argument")

    arg_names = yaml_dict.get("arg_names", [f"arg{i}" for i in range(len(args))])
    input_type = yaml_dict.get("input_type")

    if input_type == "BUILTIN":
        input_type = ops.udf.InputType.BUILTIN
    elif input_type == "PYARROW":
        input_type = ops.udf.InputType.PYARROW

    fields = {
        name: Argument(pattern=rlz.ValueOf(arg.type()), typehint=arg.type())
        for name, arg in zip(arg_names, args)
    }

    bases = (ops.ScalarUDF,)
    meta = {
        "dtype": dt.dtype(yaml_dict["type"]["name"]),
        "__input_type__": input_type,
        "__func__": property(fget=lambda _, f=fn: f),
        "__config__": {"volatility": "immutable"},
        "__udf_namespace__": None,
        "__module__": yaml_dict.get("module", "__main__"),
        "__func_name__": yaml_dict["func_name"],
    }

    kwds = {**fields, **meta}
    class_name = yaml_dict.get("class_name", yaml_dict["func_name"])

    node = type(
        class_name,
        bases,
        kwds,
    )

    return node(*args).to_expr()


@translate_to_yaml.register(FlightExpr)
def flight_expr_to_yaml(op: FlightExpr, context: any) -> dict:
    input_expr_yaml = translate_to_yaml(op.input_expr, context)
    unbound_expr_yaml = translate_to_yaml(op.unbound_expr, context)

    schema_id = context.schema_registry.register_schema(op.schema)

    make_server_pickle = serialize_udf_function(op.make_server)
    make_connection_pickle = serialize_udf_function(op.make_connection)

    return freeze(
        {
            "op": "FlightExpr",
            "name": op.name,
            "schema_ref": schema_id,
            "input_expr": input_expr_yaml,
            "unbound_expr": unbound_expr_yaml,
            "make_server": make_server_pickle,
            "make_connection": make_connection_pickle,
            "do_instrument_reader": op.do_instrument_reader,
        }
    )


@register_from_yaml_handler("FlightExpr")
def flight_expr_from_yaml(yaml_dict: Dict, context: Any) -> Any:
    name = yaml_dict.get("name")
    input_expr_yaml = yaml_dict.get("input_expr")
    unbound_expr_yaml = yaml_dict.get("unbound_expr")
    make_server_pickle = yaml_dict.get("make_server")
    make_connection_pickle = yaml_dict.get("make_connection")
    do_instrument_reader = yaml_dict.get("do_instrument_reader", False)

    input_expr = translate_from_yaml(input_expr_yaml, context)
    unbound_expr = translate_from_yaml(unbound_expr_yaml, context)

    make_server = (
        deserialize_udf_function(make_server_pickle) if make_server_pickle else None
    )
    make_connection = (
        deserialize_udf_function(make_connection_pickle)
        if make_connection_pickle
        else None
    )

    return FlightExpr.from_exprs(
        input_expr=input_expr,
        unbound_expr=unbound_expr,
        make_server=make_server,
        make_connection=make_connection,
        name=name,
        do_instrument_reader=do_instrument_reader,
    ).to_expr()


@translate_to_yaml.register(FlightUDXF)
def flight_udxf_to_yaml(op: FlightUDXF, context: any) -> dict:
    input_expr_yaml = translate_to_yaml(op.input_expr, context)
    schema_id = context.schema_registry.register_schema(op.schema)
    udxf_pickle = serialize_udf_function(op.udxf)
    make_server_pickle = serialize_udf_function(op.make_server)
    make_connection_pickle = serialize_udf_function(op.make_connection)

    return freeze(
        {
            "op": "FlightUDXF",
            "name": op.name,
            "schema_ref": schema_id,
            "input_expr": input_expr_yaml,
            "udxf": udxf_pickle,
            "make_server": make_server_pickle,
            "make_connection": make_connection_pickle,
            "do_instrument_reader": op.do_instrument_reader,
        }
    )


@register_from_yaml_handler("FlightUDXF")
def flight_udxf_from_yaml(yaml_dict: Dict, context: Any) -> Any:
    name = yaml_dict.get("name")
    input_expr_yaml = yaml_dict.get("input_expr")
    udxf_pickle = yaml_dict.get("udxf")
    make_server_pickle = yaml_dict.get("make_server")
    make_connection_pickle = yaml_dict.get("make_connection")
    do_instrument_reader = yaml_dict.get("do_instrument_reader", False)

    input_expr = translate_from_yaml(input_expr_yaml, context)
    udxf = deserialize_udf_function(udxf_pickle)
    make_server = (
        deserialize_udf_function(make_server_pickle) if make_server_pickle else None
    )
    make_connection = (
        deserialize_udf_function(make_connection_pickle)
        if make_connection_pickle
        else None
    )

    return FlightUDXF.from_expr(
        input_expr=input_expr,
        udxf=udxf,
        make_server=make_server,
        make_connection=make_connection,
        name=name,
        do_instrument_reader=do_instrument_reader,
    ).to_expr()



import pytest


SU = pytest.importorskip("xorq.common.utils.snowflake_utils")


@pytest.mark.snowflake
def test_setup_session():
    (database, schema) = ("SNOWFLAKE_SAMPLE_DATA", "TPCH_SF1")
    con = SU.make_connection(
        database=database,
        schema=schema,
        create_object_udfs=False,
    )
    dct = (
        con.raw_sql("SELECT CURRENT_WAREHOUSE(), CURRENT_DATABASE(), CURRENT_SCHEMA();")
        .fetch_pandas_all()
        .iloc[0]
        .to_dict()
    )
    assert con.current_catalog == f"{database}/{schema}"
    assert con.current_database is None
    assert con.con.database == f"{database}/{schema}"
    assert con.con.schema is None
    assert dct == {
        "CURRENT_WAREHOUSE()": "COMPUTE_WH",
        "CURRENT_DATABASE()": None,
        "CURRENT_SCHEMA()": None,
    }

    con = SU.make_connection(
        database=database,
        schema=schema,
        create_object_udfs=True,
    )
    dct = (
        con.raw_sql("SELECT CURRENT_WAREHOUSE(), CURRENT_DATABASE(), CURRENT_SCHEMA();")
        .fetch_pandas_all()
        .iloc[0]
        .to_dict()
    )
    assert con.current_catalog == database
    assert con.current_database == schema
    assert con.con.database == database
    assert con.con.schema == schema
    assert dct == {
        "CURRENT_WAREHOUSE()": "COMPUTE_WH",
        "CURRENT_DATABASE()": database,
        "CURRENT_SCHEMA()": schema,
    }


@pytest.mark.snowflake
def test_table_namespace():
    (database, schema, table_name) = ("SNOWFLAKE_SAMPLE_DATA", "TPCH_SF1", "CUSTOMER")
    con = SU.make_connection(
        database=database,
        schema=schema,
    )
    table = con.table(table_name)
    namespace = table.op().namespace
    assert namespace.catalog is not None
    assert namespace.database is not None



import os


def make_s3_credentials_defaults():
    return {
        "aws.access_key_id": os.environ.get("AWS_ACCESS_KEY_ID"),
        "aws.secret_access_key": os.environ.get("AWS_SECRET_ACCESS_KEY"),
    }


def make_s3_connection():
    connection = {
        **make_s3_credentials_defaults(),
        "aws.session_token": os.environ.get("AWS_SESSION_TOKEN", ""),
        "aws.allow_http": os.environ.get("AWS_ALLOW_HTTP", "false"),
    }

    if region := os.environ.get("AWS_REGION"):
        connection["aws.region"] = region

    return connection, connection_is_set(connection)


def connection_is_set(connection: dict[str, str]):
    keys = ("aws.access_key_id", "aws.secret_access_key")
    return all(connection[value] for value in keys)



import argparse
import random
import time
import traceback
from datetime import datetime
from functools import partial
from pathlib import Path

import pyarrow as pa

import xorq as xo
from xorq.flight import (
    FlightServer,
    FlightUrl,
)
from xorq.flight.client import (
    FlightClient,
)


db_path = "multi_duck.db"
port = 8816
table_name = "concurrent_test"


def read_data(expr, client):
    result = client.execute_query(expr)
    try:
        ((count,),) = result.to_pandas().values
        print(f"{datetime.now().isoformat()} count: {count}")
    except Exception:
        print(f"result: {result}")
        traceback.print_exc()


def write_data(table_name, client):
    data = pa.Table.from_pylist(
        (
            {
                "id": int(time.time()),
                "value": f"val-{random.randint(100, 999)}",
            },
        )
    )
    client.upload_data(table_name, data)
    print(f"{datetime.now().isoformat()} - Uploaded data: {data.to_pydict()}")


def run_server(db_path, table_name, port):
    db_path = Path(db_path).absolute()
    db_path.unlink(missing_ok=True)
    flight_server = FlightServer(
        FlightUrl(port=port),
        connection=partial(xo.duckdb.connect, db_path),
    )
    flight_server.serve()
    flight_server.server._conn.create_table(
        table_name, schema=xo.schema({"id": int, "value": str})
    )
    print(f"DuckDB Flight server started at grpc://localhost:{port}")
    while flight_server.server is not None:
        time.sleep(1)


def run_reader(table_name, port):
    expr = xo.table({"id": int}, name=table_name).count()
    client = FlightClient(port=port)
    while True:
        read_data(expr, client)
        time.sleep(1)


def run_writer(table_name, port):
    client = FlightClient(port=port)
    while True:
        write_data(table_name, client)
        time.sleep(1)


def parse_args(override=None):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "command",
        choices=("serve", "read", "write"),
    )
    parser.add_argument(
        "-d",
        "--db-path",
        default=db_path,
    )
    parser.add_argument(
        "-p",
        "--port",
        default=port,
        type=int,
    )
    parser.add_argument(
        "-n",
        "--table-name",
        default=table_name,
    )
    args = parser.parse_args(override)
    return args


if __name__ == "__main__":
    args = parse_args()
    if args.command == "serve":
        run_server(args.db_path, args.table_name, args.port)
    elif args.command == "read":
        run_reader(args.table_name, args.port)
    elif args.command == "write":
        run_writer(args.table_name, args.port)
    else:
        raise ValueError



import xorq.vendor.ibis as ibis


def test_string_concat(compiler):
    s1 = ibis.literal("hello")
    s2 = ibis.literal("world")
    expr = s1 + s2
    yaml_dict = compiler.to_yaml(expr)["expression"]

    assert yaml_dict["op"] == "StringConcat"
    assert yaml_dict["args"][0]["value"] == "hello"
    assert yaml_dict["args"][1]["value"] == "world"
    assert yaml_dict["type"] == {"name": "String", "nullable": True}


def test_string_upper_lower(compiler):
    s = ibis.literal("Hello")
    upper_expr = s.upper()
    lower_expr = s.lower()

    upper_yaml = compiler.to_yaml(upper_expr)["expression"]
    assert upper_yaml["op"] == "Uppercase"
    assert upper_yaml["args"][0]["value"] == "Hello"

    lower_yaml = compiler.to_yaml(lower_expr)["expression"]
    assert lower_yaml["op"] == "Lowercase"
    assert lower_yaml["args"][0]["value"] == "Hello"


def test_string_length(compiler):
    s = ibis.literal("hello")
    expr = s.length()
    yaml_dict = compiler.to_yaml(expr)["expression"]

    assert yaml_dict["op"] == "StringLength"
    assert yaml_dict["args"][0]["value"] == "hello"
    assert yaml_dict["type"] == {"name": "Int32", "nullable": True}


def test_string_substring(compiler):
    s = ibis.literal("hello world")
    expr = s.substr(0, 5)
    yaml_dict = compiler.to_yaml(expr)["expression"]

    assert yaml_dict["op"] == "Substring"
    assert yaml_dict["args"][0]["value"] == "hello world"
    assert yaml_dict["args"][1]["value"] == 0
    assert yaml_dict["args"][2]["value"] == 5



import xorq as xo


con = xo.connect()
iris = xo.examples.iris.fetch(backend=con, table_name="iris")
res = (
    iris.filter([iris.sepal_length > 5])
    .group_by("species")
    .agg(iris.sepal_width.sum())
    .execute()
)

print(res)



from pathlib import Path

import xorq as xo
from xorq.caching import ParquetStorage


pg = xo.postgres.connect_examples()
con = xo.connect()  # empty connection
storage = ParquetStorage(
    source=con,
    path=Path.cwd(),
)


alltypes = con.register(
    pg.table("functional_alltypes"), table_name="pg_functional_alltypes"
)
cached = (
    alltypes.select(alltypes.smallint_col, alltypes.int_col, alltypes.float_col).cache(
        storage=storage
    )  # cache expression (this creates a local table)
)
expr = cached.filter(
    [
        cached.float_col > 0,
        cached.smallint_col > 4,
        cached.int_col < cached.float_col * 2,
    ]
)
path = storage.get_loc(cached.ls.get_key())


print(f"{path} exists?: {path.exists()}")
result = xo.execute(cached)  # the filter is executed on the local table
print(f"{path} exists?: {path.exists()}")
print(result)



import pickle

import pandas as pd
import toolz
import xgboost as xgb

import xorq as xo
import xorq.expr.datatypes as dt
import xorq.expr.udf as udf
from xorq.expr.udf import (
    make_pandas_expr_udf,
)


ROWNUM = "rownum"
features = (
    "emp_length",
    "dti",
    "annual_inc",
    "loan_amnt",
    "fico_range_high",
    "cr_age_days",
)
target = "event_occurred"
model_key = "model"
prediction_key = "predicted"
prediction_typ = "float32"


@toolz.curry
def train_xgboost_model(df, features=features, target=target, seed=0):
    param = {"max_depth": 4, "eta": 1, "objective": "binary:logistic", "seed": seed}
    num_round = 10
    if ROWNUM in df:
        # enforce order for reproducibility
        df = df.sort_values(ROWNUM, ignore_index=True)
    X = df[list(features)]
    y = df[target]
    dtrain = xgb.DMatrix(X, y)
    bst = xgb.train(param, dtrain, num_boost_round=num_round)
    return bst


@toolz.curry
def predict_xgboost_model(model, df, features=features):
    return model.predict(xgb.DMatrix(df[list(features)]))


t = xo.read_parquet(xo.config.options.pins.get_path("lending-club"))
(train, test) = xo.train_test_splits(
    t,
    unique_key=ROWNUM,
    test_sizes=0.7,
)

# manual run
model = train_xgboost_model(train.execute())
from_pd = test.execute().assign(
    **{prediction_key: predict_xgboost_model(model)},
)

# using expr-scalar-udf
model_udaf = udf.agg.pandas_df(
    fn=toolz.compose(pickle.dumps, train_xgboost_model),
    schema=t[features + (target,)].schema(),
    return_type=dt.binary,
    name=model_key,
)
predict_expr_udf = make_pandas_expr_udf(
    computed_kwargs_expr=model_udaf.on_expr(train),
    fn=predict_xgboost_model,
    schema=t[features].schema(),
    return_type=dt.dtype(prediction_typ),
    name=prediction_key,
)
from_ls = test.mutate(predict_expr_udf.on_expr(test).name(prediction_key)).execute()

pd._testing.assert_frame_equal(from_ls, from_pd)



import xorq.vendor.ibis as ibis


def test_add(compiler):
    lit1 = ibis.literal(5)
    lit2 = ibis.literal(3)
    expr = lit1 + lit2

    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]
    assert expression["op"] == "Add"
    assert expression["args"][0]["op"] == "Literal"
    assert expression["args"][0]["value"] == 5
    assert expression["args"][1]["op"] == "Literal"
    assert expression["args"][1]["value"] == 3
    assert expression["type"] == {"name": "Int8", "nullable": True}

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)


def test_subtract(compiler):
    lit1 = ibis.literal(5)
    lit2 = ibis.literal(3)
    expr = lit1 - lit2

    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]
    assert expression["op"] == "Subtract"
    assert expression["args"][0]["op"] == "Literal"
    assert expression["args"][0]["value"] == 5
    assert expression["args"][1]["op"] == "Literal"
    assert expression["args"][1]["value"] == 3
    assert expression["type"] == {"name": "Int8", "nullable": True}

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)


def test_multiply(compiler):
    lit1 = ibis.literal(5)
    lit2 = ibis.literal(3)
    expr = lit1 * lit2

    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]
    assert expression["op"] == "Multiply"
    assert expression["args"][0]["op"] == "Literal"
    assert expression["args"][0]["value"] == 5
    assert expression["args"][1]["op"] == "Literal"
    assert expression["args"][1]["value"] == 3
    assert expression["type"] == {"name": "Int8", "nullable": True}

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)


def test_divide(compiler):
    lit1 = ibis.literal(6.0)
    lit2 = ibis.literal(2.0)
    expr = lit1 / lit2

    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]
    assert expression["op"] == "Divide"
    assert expression["args"][0]["op"] == "Literal"
    assert expression["args"][0]["value"] == 6.0
    assert expression["args"][1]["op"] == "Literal"
    assert expression["args"][1]["value"] == 2.0
    assert expression["type"] == {"name": "Float64", "nullable": True}

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)


def test_mixed_arithmetic(compiler):
    i = ibis.literal(5)
    f = ibis.literal(2.5)
    expr = i * f

    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]
    assert expression["op"] == "Multiply"
    assert expression["type"] == {"name": "Float64", "nullable": True}

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)


def test_complex_arithmetic(compiler):
    a = ibis.literal(10)
    b = ibis.literal(5)
    c = ibis.literal(2.0)
    expr = (a + b) * c

    yaml_dict = compiler.to_yaml(expr)
    expression = yaml_dict["expression"]

    assert expression["op"] == "Multiply"
    assert expression["args"][0]["op"] == "Add"

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(expr)



import functools
import itertools
from typing import Any

import attr
from dask.base import tokenize

import xorq.expr.datatypes as dt
import xorq.vendor.ibis.expr.types as ir
from xorq.ibis_yaml.utils import freeze
from xorq.vendor.ibis.common.collections import FrozenOrderedDict


FROM_YAML_HANDLERS: dict[str, Any] = {}


class SchemaRegistry:
    def __init__(self):
        self.schemas = {}
        self.counter = itertools.count()
        self.nodes = {}

    def register_schema(self, schema):
        frozen_schema = freeze(
            {name: _translate_type(dtype) for name, dtype in schema.items()}
        )
        for schema_id, existing_schema in self.schemas.items():
            if existing_schema == frozen_schema:
                return schema_id
        schema_id = f"schema_{next(self.counter)}"
        self.schemas[schema_id] = frozen_schema
        return schema_id

    def _register_expr_schema(self, expr: ir.Expr) -> str:
        if hasattr(expr, "schema"):
            schema = expr.schema()
            return self.register_schema(schema)
        return None

    def register_node(self, node_dict):
        frozen_node = freeze(node_dict)

        node_hash = tokenize(frozen_node)

        if node_hash not in self.nodes:
            self.nodes[node_hash] = frozen_node

        return node_hash


@attr.s(frozen=True)
class TranslationContext:
    schema_registry: SchemaRegistry = attr.ib(factory=SchemaRegistry)
    profiles: FrozenOrderedDict = attr.ib(factory=FrozenOrderedDict)
    definitions: FrozenOrderedDict = attr.ib(
        factory=lambda: freeze({"schemas": {}, "nodes": {}})
    )

    def update_definitions(self, new_definitions: FrozenOrderedDict):
        return attr.evolve(self, definitions=new_definitions)

    def finalize_definitions(self):
        updated_defs = dict(self.definitions)
        updated_defs["schemas"] = self.schema_registry.schemas
        updated_defs["nodes"] = self.schema_registry.nodes
        return attr.evolve(self, definitions=freeze(updated_defs))


def register_from_yaml_handler(*op_names: str):
    def decorator(func):
        for name in op_names:
            FROM_YAML_HANDLERS[name] = func
        return func

    return decorator


@functools.cache
@functools.singledispatch
def translate_from_yaml(yaml_dict: dict, context: TranslationContext) -> Any:
    if "node_ref" in yaml_dict:
        node_ref = yaml_dict["node_ref"]
        if "nodes" not in context.definitions:
            raise ValueError(f"Missing 'nodes' in definitions for reference {node_ref}")

        try:
            node_dict = context.definitions["nodes"][node_ref]
            return translate_from_yaml(node_dict, context)
        except KeyError:
            raise ValueError(f"Node reference {node_ref} not found in definitions")

    op_type = yaml_dict["op"]
    if op_type not in FROM_YAML_HANDLERS:
        raise NotImplementedError(f"No handler for operation {op_type}")
    return FROM_YAML_HANDLERS[op_type](yaml_dict, context)


@functools.cache
@functools.singledispatch
def translate_to_yaml(op: Any, context: TranslationContext) -> dict:
    raise NotImplementedError(f"No translation rule for {type(op)}")


@functools.singledispatch
def _translate_type(dtype: dt.DataType) -> dict:
    return freeze({"name": type(dtype).__name__, "nullable": dtype.nullable})



import datetime

import pandas as pd
import pyarrow as pa
import pytest

import xorq as xo
import xorq.flight.action as A
import xorq.flight.exchanger as E
from xorq.common.utils.rbr_utils import instrument_reader
from xorq.flight import FlightServer, FlightUrl
from xorq.flight.action import AddExchangeAction
from xorq.flight.exchanger import EchoExchanger, PandasUDFExchanger


def make_flight_url(port):
    if port is not None:
        assert not FlightUrl.port_in_use(port), f"Port {port} already in use"
    flight_url = FlightUrl(port=port)
    assert FlightUrl.port_in_use(flight_url.port), (
        f"Port {flight_url.port} should be in use"
    )
    return flight_url


@pytest.mark.parametrize(
    "connection,port",
    [
        pytest.param(xo.duckdb.connect, 5005, id="duckdb"),
        pytest.param(xo.datafusion.connect, 5005, id="datafusion"),
        pytest.param(xo.connect, 5005, id="xorq"),
    ],
)
def test_port_in_use(connection, port):
    assert port is not None
    flight_url = make_flight_url(port)
    with pytest.raises(OSError, match="Address already in use"):
        with FlightServer(
            flight_url=flight_url,
            connection=connection,
        ) as _:
            # entering the above context releases the port
            # so we won't raise until we enter the second context and try to use it
            flight_url2 = FlightUrl(port=port)  # noqa: F841


class Answer42Action(A.AbstractAction):
    @classmethod
    @property
    def name(cls):
        return "answer-42"

    @classmethod
    def description(cls):
        return (
            "the answer to the ultimate question of life, the universe, and everything"
        )

    @classmethod
    def do_action(cls, server, context, action):
        yield A.make_flight_result(42)


def test_list_actions():
    with FlightServer() as flight_server:
        actions = flight_server.client.do_action_one(A.ListActionsAction.name)
        assert actions == tuple(A.actions)


def test_add_action():
    with FlightServer() as flight_server:
        actions = flight_server.client.do_action_one(A.ListActionsAction.name)
        assert Answer42Action.name not in actions
        flight_server.client.do_action(A.AddActionAction.name, Answer42Action)
        actions = flight_server.client.do_action_one(A.ListActionsAction.name)
        assert Answer42Action.name in actions


def test_list_exchanges():
    with FlightServer() as flight_server:
        exchanges = flight_server.client.do_action_one(A.ListExchangesAction.name)
        assert exchanges == tuple(E.exchangers)


@pytest.mark.parametrize(
    "connection,port",
    [
        pytest.param(xo.duckdb.connect, None, id="duckdb"),
        pytest.param(xo.datafusion.connect, None, id="datafusion"),
        pytest.param(xo.connect, None, id="xorq"),
    ],
)
def test_register_and_list_tables(connection, port):
    flight_url = make_flight_url(port)

    with FlightServer(
        flight_url=flight_url,
        verify_client=False,
        connection=connection,
    ) as main:
        con = main.con
        assert con.version is not None

        data = pa.table(
            {"id": [1, 2, 3], "name": ["Alice", "Bob", "Charlie"]}
        ).to_pandas()

        con.register(data, table_name="users")
        t = con.table("users")
        actual = xo.execute(t)

        assert t.schema() is not None
        assert "users" in con.list_tables()
        assert isinstance(actual, pd.DataFrame)


@pytest.mark.parametrize(
    "connection,port",
    [
        pytest.param(xo.duckdb.connect, None, id="duckdb"),
        pytest.param(xo.datafusion.connect, None, id="datafusion"),
        pytest.param(xo.connect, None, id="xorq"),
    ],
)
def test_into_backend_flight_server(connection, port, parquet_dir):
    batting = xo.read_parquet(parquet_dir / "batting.parquet")
    flight_url = make_flight_url(port)

    with FlightServer(
        flight_url=flight_url,
        verify_client=False,
        connection=connection,
    ) as main:
        con = main.con
        t = batting.filter(batting.yearID == 2015).into_backend(con, "xo_batting")
        expr = (
            t.join(t, "playerID")
            .limit(15)
            .select(player_id="playerID", year_id="yearID_right")
        )

        assert not expr.execute().empty


@pytest.mark.parametrize(
    "connection,port",
    [
        pytest.param(xo.duckdb.connect, None, id="duckdb"),
        pytest.param(xo.datafusion.connect, None, id="datafusion"),
        pytest.param(xo.connect, None, id="xorq"),
    ],
)
def test_read_parquet(connection, port, parquet_dir):
    flight_url = make_flight_url(port)
    with FlightServer(
        flight_url=flight_url,
        verify_client=False,
        connection=connection,
    ) as main:
        con = main.con
        batting = con.read_parquet(parquet_dir / "batting.parquet")
        assert xo.execute(batting) is not None


@pytest.mark.parametrize(
    "connection,port",
    [
        pytest.param(xo.duckdb.connect, None, id="duckdb"),
        pytest.param(xo.datafusion.connect, None, id="datafusion"),
        pytest.param(xo.connect, None, id="xorq"),
    ],
)
def test_exchange(connection, port):
    flight_url = make_flight_url(port)

    def my_f(df):
        return df[["a", "b"]].sum(axis=1)

    with FlightServer(
        flight_url=flight_url,
        verify_client=False,
        connection=connection,
    ) as main:
        client = main.client
        udf_exchanger = PandasUDFExchanger(
            my_f,
            schema_in=pa.schema(
                (
                    pa.field("a", pa.int64()),
                    pa.field("b", pa.int64()),
                )
            ),
            name="x",
            typ=pa.int64(),
            append=True,
        )
        client.do_action(AddExchangeAction.name, udf_exchanger, options=client._options)

        # a small example
        df_in = pd.DataFrame({"a": [1], "b": [2], "c": [100]})
        fut, rbr = client.do_exchange(
            udf_exchanger.command,
            pa.RecordBatchReader.from_stream(df_in),
        )
        df_out = rbr.read_pandas()
        writes_reads = fut.result()

        assert writes_reads["n_writes"] == writes_reads["n_reads"]
        assert df_out is not None

        # demonstrate streaming
        df_in = pd.DataFrame(
            {
                "a": range(100_000),
                "b": range(100_000, 200_000),
                "c": range(200_000, 300_000),
            }
        )
        fut, rbr = client.do_exchange_batches(
            udf_exchanger.command,
            instrument_reader(pa.Table.from_pandas(df_in).to_reader(max_chunksize=100)),
        )
        first_batch = next(rbr)
        first_batch_time = datetime.datetime.now()
        assert first_batch is not None, f"must get first batch by {first_batch_time}"

        rest = rbr.read_pandas()
        rest_time = datetime.datetime.now()
        assert rest is not None, f"must get first batch by {rest_time}"
        assert first_batch_time < rest_time

        writes_reads = fut.result()
        assert writes_reads["n_writes"] == 1000  # because
        assert writes_reads["n_reads"] == 1000


@pytest.mark.parametrize(
    "connection",
    (
        xo.duckdb.connect,
        xo.datafusion.connect,
        xo.connect,
    ),
)
def test_reentry(connection):
    df_in = pd.DataFrame({"a": [1], "b": [2], "c": [100]})
    with FlightServer(
        verify_client=False,
        connection=connection,
    ) as server:
        fut, rbr = server.client.do_exchange(
            EchoExchanger.command,
            pa.RecordBatchReader.from_stream(df_in),
        )
        df_out = rbr.read_pandas()
        assert df_in.equals(df_out)
    with server:
        fut, rbr = server.client.do_exchange(
            EchoExchanger.command,
            pa.RecordBatchReader.from_stream(df_in),
        )
        df_out = rbr.read_pandas()
        assert df_in.equals(df_out)


@pytest.mark.parametrize(
    "connection",
    (
        xo.duckdb.connect,
        xo.datafusion.connect,
        xo.connect,
    ),
)
def test_serve_close(connection):
    df_in = pd.DataFrame({"a": [1], "b": [2], "c": [100]})
    server = FlightServer(
        verify_client=False,
        connection=connection,
    )

    server.serve()
    fut, rbr = server.client.do_exchange(
        EchoExchanger.command,
        pa.RecordBatchReader.from_stream(df_in),
    )
    df_out = rbr.read_pandas()
    assert df_in.equals(df_out)
    server.close()

    server.serve()
    fut, rbr = server.client.do_exchange(
        EchoExchanger.command,
        pa.RecordBatchReader.from_stream(df_in),
    )
    df_out = rbr.read_pandas()
    assert df_in.equals(df_out)
    server.close()


def test_ctor_exchanger_registration():
    def dummy(df: pd.DataFrame):
        return pd.DataFrame({"row_count": [42]})

    schema_in = xo.schema({"dummy": "int64"})
    dummy_udxf = E.make_udxf(
        dummy,
        schema_in.to_pyarrow(),
        xo.schema({"row_count": "int64"}).to_pyarrow(),
    )
    flight_server = FlightServer(exchangers=[dummy_udxf])
    with flight_server:
        client = flight_server.client
        available = client.do_action_one(A.ListExchangesAction.name)
        assert dummy_udxf.command in available
        client.do_exchange(
            dummy_udxf.command,
            xo.memtable({"dummy": [0]}, schema=schema_in).to_pyarrow_batches(),
        )


@pytest.mark.parametrize(
    "expr",
    [
        pytest.param(xo.table({"id": int}, name="users").count(), id="scalar"),
        pytest.param(xo.literal(1), id="literal"),
        pytest.param(xo.table({"id": int}, name="users").id, id="column"),
    ],
)
def test_execute_query_non_relation_expr(expr):
    flight_url = make_flight_url(None)
    data = pa.table({"id": [1, 2, 3], "name": ["Alice", "Bob", "Charlie"]}).to_pandas()

    with FlightServer(
        flight_url=flight_url,
        verify_client=False,
        connection=xo.duckdb.connect,
    ) as main:
        main.con.register(data, table_name="users")
        actual = main.client.execute_query(expr)
        assert isinstance(actual, pa.Table)
        assert len(actual) > 0



import hashlib
import operator
import pathlib
import re

import dask
import pytest
import toolz

import xorq as xo
import xorq.common.utils.dask_normalize  # noqa: F401
from xorq.caching import (
    SourceSnapshotStorage,
)
from xorq.common.utils.dask_normalize.dask_normalize_utils import (
    patch_normalize_token,
    walk_normalized,
)


def test_ensure_deterministic():
    assert dask.config.get("tokenize.ensure-deterministic")


def test_unregistered_raises():
    class Unregistered:
        pass

    with pytest.raises(ValueError, match="cannot be deterministically hashed"):
        dask.base.tokenize(Unregistered())


@pytest.mark.snapshot_check
def test_tokenize_datafusion_memory_expr(alltypes_df, snapshot):
    con = xo.datafusion.connect()
    typ = type(con)
    t = con.register(alltypes_df, "t")
    with patch_normalize_token(type(con)) as mocks:
        actual = dask.base.tokenize(t)
    mocks[typ].assert_not_called()
    snapshot.assert_match(actual, "datafusion_memory_key.txt")


@pytest.mark.snapshot_check
def test_tokenize_datafusion_parquet_expr(alltypes_df, tmp_path, snapshot):
    path = pathlib.Path(tmp_path).joinpath("data.parquet")
    alltypes_df.to_parquet(path)
    con = xo.datafusion.connect()
    t = con.register(path, "t")
    # work around tmp_path variation
    (prefix, suffix) = (
        re.escape(part)
        for part in (
            r"file_groups={1 group: [[",
            r"]]",
        )
    )
    to_hash = re.sub(
        prefix + f".*?/{path.name}" + suffix,
        prefix + f"/{path.name}" + suffix,
        str(tuple(dask.base.normalize_token(t))),
    )
    actual = hashlib.md5(to_hash.encode(), usedforsecurity=False).hexdigest()
    snapshot.assert_match(actual, "datafusion_key.txt")


@pytest.mark.snapshot_check
def test_tokenize_pandas_expr(alltypes_df, snapshot):
    con = xo.pandas.connect()
    typ = type(con)
    t = con.create_table("t", alltypes_df)
    with patch_normalize_token(type(t.op().source)) as mocks:
        actual = dask.base.tokenize(t)
    mocks[typ].assert_not_called()
    snapshot.assert_match(actual, "pandas_key.txt")


@pytest.mark.snapshot_check
def test_tokenize_duckdb_expr(batting, snapshot):
    con = xo.duckdb.connect()
    typ = type(con)
    t = con.register(batting.to_pyarrow(), "dashed-name")
    with patch_normalize_token(type(con)) as mocks:
        actual = dask.base.tokenize(t)
    mocks[typ].assert_not_called()

    snapshot.assert_match(actual, "duckdb_key.txt")


@pytest.mark.snapshot_check
def test_pandas_snapshot_key(alltypes_df, snapshot):
    con = xo.pandas.connect()
    t = con.create_table("t", alltypes_df)
    storage = SourceSnapshotStorage(source=con)
    actual = storage.get_key(t)
    snapshot.assert_match(actual, "pandas_snapshot_key.txt")


@pytest.mark.snapshot_check
def test_duckdb_snapshot_key(batting, snapshot):
    con = xo.duckdb.connect()
    t = con.register(batting.to_pyarrow(), "dashed-name")
    storage = SourceSnapshotStorage(source=con)
    actual = storage.get_key(t)
    snapshot.assert_match(actual, "duckdb_snapshot_key.txt")


@pytest.mark.parametrize(
    "target, expected",
    (
        (1, 1),
        (1.0, 1),
        ("two", 2),
        # bytes are converted to text
        (b"three", 0),
        ("three", 3),
    ),
)
def test_walk_normalized(target, expected):
    normalized = (
        [
            (
                b"three",
                b"three",
                1,
                [
                    "two",
                ],
            ),
        ],
        (
            2,
            (
                3,
                b"three",
            ),
            "two",
        ),
    )
    f = toolz.curry(operator.eq, target)
    actual = sum(walk_normalized(f, normalized))
    assert actual == expected



import pandas as pd
import pyarrow as pa
from pandas.testing import assert_series_equal

import xorq as xo
from xorq.expr.udf import pyarrow_udwf
from xorq.internal import WindowEvaluator
from xorq.vendor import ibis


class ExponentialSmoothDefault(WindowEvaluator):
    """Create a running smooth operation across an entire partition at once."""

    def __init__(self, alpha: float) -> None:
        self.alpha = alpha

    def evaluate_all(self, values: list[pa.Array], num_rows: int) -> pa.Array:
        results = []
        curr_value = 0.0
        values = values[0]
        for idx in range(num_rows):
            if idx == 0:
                curr_value = values[idx].as_py()
            else:
                curr_value = values[idx].as_py() * self.alpha + curr_value * (
                    1.0 - self.alpha
                )
            results.append(curr_value)

        return pa.array(results)


class SmoothBoundedFromPreviousRow(WindowEvaluator):
    """Smooth over from the previous to current row only."""

    def __init__(self, alpha: float) -> None:
        self.alpha = alpha

    def supports_bounded_execution(self) -> bool:
        return True

    def get_range(self, idx: int, num_rows: int) -> tuple[int, int]:
        # Override the default range of current row since uses_window_frame is False
        # So for the purpose of this test we just smooth from the previous row to
        # current.
        if idx == 0:
            return (0, 0)
        return (idx - 1, idx)

    def evaluate(
        self, values: list[pa.Array], eval_range: tuple[int, int]
    ) -> pa.Scalar:
        (start, stop) = eval_range
        curr_value = 0.0
        values = values[0]
        for idx in range(start, stop + 1):
            if idx == start:
                curr_value = values[idx].as_py()
            else:
                curr_value = values[idx].as_py() * self.alpha + curr_value * (
                    1.0 - self.alpha
                )
        return pa.scalar(curr_value).cast(pa.float64())


class SmoothAcrossRank(WindowEvaluator):
    """Smooth over the rank from the previous rank to current."""

    def __init__(self, alpha: float) -> None:
        self.alpha = alpha

    def include_rank(self) -> bool:
        return True

    def evaluate_all_with_rank(
        self, num_rows: int, ranks_in_partition: list[tuple[int, int]]
    ) -> pa.Array:
        results = []
        for idx in range(num_rows):
            if idx == 0:
                prior_value = 1.0
            matching_row = [
                i
                for i in range(len(ranks_in_partition))
                if ranks_in_partition[i][0] <= idx and ranks_in_partition[i][1] > idx
            ][0] + 1
            curr_value = matching_row * self.alpha + prior_value * (1.0 - self.alpha)
            results.append(curr_value)
            prior_value = matching_row

        return pa.array(results)


class ExponentialSmoothFrame(WindowEvaluator):
    "Find the value across an entire frame using exponential smoothing"

    def __init__(self, alpha: float) -> None:
        self.alpha = alpha

    def uses_window_frame(self) -> bool:
        return True

    def evaluate(
        self, values: list[pa.Array], eval_range: tuple[int, int]
    ) -> pa.Scalar:
        (start, stop) = eval_range
        curr_value = 0.0
        if len(values) > 1:
            order_by = values[1]  # noqa: F841
            values = values[0]
        else:
            values = values[0]
        for idx in range(start, stop):
            if idx == start:
                curr_value = values[idx].as_py()
            else:
                curr_value = values[idx].as_py() * self.alpha + curr_value * (
                    1.0 - self.alpha
                )
        return pa.scalar(curr_value).cast(pa.float64())


class SmoothTwoColumn(WindowEvaluator):
    """Smooth once column based on a condition of another column.

    If the second column is above a threshold, then smooth over the first column from
    the previous and next rows.
    """

    def __init__(self, alpha: float) -> None:
        self.alpha = alpha

    def evaluate_all(self, values: list[pa.Array], num_rows: int) -> pa.Array:
        results = []
        values_a = values[0]
        values_b = values[1]
        for idx in range(num_rows):
            if not values_b[idx].is_valid:
                if idx == 0:
                    results.append(values_a[1].cast(pa.float64()))
                elif idx == num_rows - 1:
                    results.append(values_a[num_rows - 2].cast(pa.float64()))
                else:
                    results.append(
                        pa.scalar(
                            values_a[idx - 1].as_py() * self.alpha
                            + values_a[idx + 1].as_py() * (1.0 - self.alpha)
                        )
                    )
            else:
                results.append(values_a[idx].cast(pa.float64()))

        return pa.array(results)


@pyarrow_udwf(
    schema=ibis.schema({"a": float}),
    return_type=ibis.dtype(float),
    alpha=0.9,
)
def exp_smooth(self, values: list[pa.Array], num_rows: int) -> pa.Array:
    results = []
    curr_value = 0.0
    values = values[0]
    for idx in range(num_rows):
        if idx == 0:
            curr_value = values[idx].as_py()
        else:
            curr_value = values[idx].as_py() * self.alpha + curr_value * (
                1.0 - self.alpha
            )
        results.append(curr_value)

    return pa.array(results)


@pyarrow_udwf(
    schema=ibis.schema({"a": float}),
    return_type=ibis.dtype(float),
    supports_bounded_execution=True,
    get_range=SmoothBoundedFromPreviousRow.get_range,
    alpha=0.9,
)
def smooth_two_row(
    self, values: list[pa.Array], eval_range: tuple[int, int]
) -> pa.Scalar:
    (start, stop) = eval_range
    curr_value = 0.0
    values = values[0]
    for idx in range(start, stop + 1):
        if idx == start:
            curr_value = values[idx].as_py()
        else:
            curr_value = values[idx].as_py() * self.alpha + curr_value * (
                1.0 - self.alpha
            )
    return pa.scalar(curr_value).cast(pa.float64())


@pyarrow_udwf(
    schema=ibis.schema({"a": float}),
    return_type=ibis.dtype(float),
    alpha=0.9,
    include_rank=True,
)
def smooth_rank(
    self, num_rows: int, ranks_in_partition: list[tuple[int, int]]
) -> pa.Array:
    results = []
    for idx in range(num_rows):
        if idx == 0:
            prior_value = 1.0
        matching_row = [
            i
            for i in range(len(ranks_in_partition))
            if ranks_in_partition[i][0] <= idx and ranks_in_partition[i][1] > idx
        ][0] + 1
        curr_value = matching_row * self.alpha + prior_value * (1.0 - self.alpha)
        results.append(curr_value)
        prior_value = matching_row

    return pa.array(results)


@pyarrow_udwf(
    schema=ibis.schema({"a": float}),
    return_type=ibis.dtype(float),
    alpha=0.9,
    uses_window_frame=True,
)
def smooth_frame(
    self, values: list[pa.Array], eval_range: tuple[int, int]
) -> pa.Scalar:
    (start, stop) = eval_range
    curr_value = 0.0
    if len(values) > 1:
        order_by = values[1]  # noqa: F841
        values = values[0]
    else:
        values = values[0]
    for idx in range(start, stop):
        if idx == start:
            curr_value = values[idx].as_py()
        else:
            curr_value = values[idx].as_py() * self.alpha + curr_value * (
                1.0 - self.alpha
            )
    return pa.scalar(curr_value).cast(pa.float64())


@pyarrow_udwf(
    schema=ibis.schema({"a": float, "b": int}),
    return_type=ibis.dtype(float),
    alpha=0.9,
)
def smooth_two_col(self, values: list[pa.Array], num_rows: int) -> pa.Array:
    results = []
    values_a = values[0]
    values_b = values[1]
    for idx in range(num_rows):
        if not values_b[idx].is_valid:
            if idx == 0:
                results.append(values_a[1].cast(pa.float64()))
            elif idx == num_rows - 1:
                results.append(values_a[num_rows - 2].cast(pa.float64()))
            else:
                results.append(
                    pa.scalar(
                        values_a[idx - 1].as_py() * self.alpha
                        + values_a[idx + 1].as_py() * (1.0 - self.alpha)
                    )
                )
        else:
            results.append(values_a[idx].cast(pa.float64()))

    return pa.array(results)


con = xo.connect()
t = con.register(
    pa.Table.from_batches(
        [
            pa.RecordBatch.from_arrays(
                [
                    pa.array([1.0, 2.1, 2.9, 4.0, 5.1, 6.0, 6.9, 8.0]),
                    pa.array([1, 2, None, 4, 5, 6, None, 8]),
                    pa.array(["A", "A", "A", "A", "A", "B", "B", "B"]),
                ],
                names=["a", "b", "c"],
            )
        ]
    ),
    table_name="t",
)
expr = (
    t.mutate(exp_smooth=exp_smooth.on_expr(t).round(3))
    .mutate(smooth_two_row=smooth_two_row.on_expr(t))
    .mutate(smooth_rank=smooth_rank.on_expr(t).over(ibis.window(order_by="c")))
    .mutate(
        smooth_two_col=smooth_two_col.on_expr(t).over(
            ibis.window(preceding=None, following=0)
        )
    )
    .mutate(
        smooth_frame=smooth_frame.on_expr(t)
        .over(ibis.window(preceding=None, following=0))
        .round(3)
    )
)
result = xo.execute(expr).sort_values(["c", "a"], ignore_index=True)
expected = (
    t.execute()
    .combine_first(
        pd.DataFrame(
            {
                "exp_smooth": pa.array(
                    [1, 1.99, 2.809, 3.881, 4.978, 5.898, 6.8, 7.88]
                ),
                "smooth_two_col": pa.array([1, 2.1, 2.29, 4, 5.1, 6, 6.2, 8.0]),
                "smooth_two_row": pa.array(
                    [1.0, 1.99, 2.82, 3.89, 4.99, 5.91, 6.81, 7.89]
                ),
                "smooth_rank": pa.array([1, 1, 1, 1, 1, 1.9, 2.0, 2.0]),
                "smooth_frame": pa.array(
                    [1, 1.99, 2.809, 3.881, 4.978, 5.898, 6.8, 7.88]
                ),
            }
        )
    )
    .reindex_like(result)
)


for col in expected.filter(like="smooth").columns:
    try:
        assert_series_equal(result[col], expected[col])
    except Exception:
        print(f"col {col} failed")



import xorq as xo
from xorq.expr.relations import into_backend


pg = xo.postgres.connect_env()
db = xo.duckdb.connect()

batting = pg.table("batting")

awards_players = xo.examples.awards_players.fetch(backend=db)
left = batting.filter(batting.yearID == 2015)
right = awards_players.filter(awards_players.lgID == "NL").drop("yearID", "lgID")
expr = left.join(into_backend(right, pg), ["playerID"], how="semi")[["yearID", "stint"]]


result = xo.execute(expr)
print(result)



"""Initialize xorq module."""

from __future__ import annotations

from xorq import examples
from xorq import udf
from xorq.udf import *  # noqa: F403
from xorq.config import options
from xorq.expr import api
from xorq.expr.api import *  # noqa: F403
from xorq import ml
from xorq.ml import *  # noqa: F403
from xorq.backends.let import Backend
from xorq.internal import SessionConfig

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata

__all__ = [  # noqa: PLE0604
    "api",
    "examples",
    "connect",
    "options",
    "SessionConfig",
    "udf",
    *api.__all__,
    *ml.__all__,
    *udf.__all__,
]

_CUSTOM_BACKENDS = ["postgres", "snowflake"]


def _load_entry_points():
    eps = importlib_metadata.entry_points(group="xorq.backends")
    return sorted(eps)


def load_backend(name):
    if entry_point := next(
        (ep for ep in _load_entry_points() if ep.name == name), None
    ):
        import types

        import xorq as xo

        module = entry_point.load()
        backend = module.Backend()
        backend.register_options()

        def connect(*args, **kwargs):
            return backend.connect(*args, **kwargs)

        connect.__doc__ = backend.do_connect.__doc__
        connect.__wrapped__ = backend.do_connect
        connect.__module__ = f"xorq.{name}"

        proxy = types.ModuleType(f"xorq.{name}")
        setattr(xo, name, proxy)
        proxy.connect = connect
        proxy.compile = backend.compile
        proxy.has_operation = backend.has_operation
        proxy.name = name
        proxy._from_url = backend._from_url
        # Add any additional methods that should be exposed at the top level
        for attr in getattr(backend, "_top_level_methods", ()):
            setattr(proxy, attr, getattr(backend, attr))

        return proxy


def connect(session_config: SessionConfig | None = None) -> Backend:
    """Create a xorq backend."""
    instance = Backend()
    instance.do_connect(session_config)
    return instance


def __getattr__(name):
    from xorq.vendor import ibis

    return load_backend(name) or ibis.load_backend(name)


__version__ = importlib_metadata.version(__package__)



import pickle

from sklearn.feature_extraction.text import TfidfVectorizer

import xorq as xo
import xorq.vendor.ibis.expr.datatypes as dt
from xorq.caching import ParquetStorage
from xorq.common.utils.defer_utils import deferred_read_parquet
from xorq.common.utils.import_utils import import_python
from xorq.ml import (
    deferred_fit_transform_series_sklearn,
    train_test_splits,
)


m = import_python(xo.options.pins.get_path("hackernews_lib"))


deferred_fit_transform_tfidf = deferred_fit_transform_series_sklearn(
    col="title", cls=TfidfVectorizer, return_type=dt.Array(dt.float64)
)


con = xo.connect()
train_expr, test_expr = (
    deferred_read_parquet(
        con,
        xo.options.pins.get_path("hn-fetcher-input-small.parquet"),
        "fetcher-input",
    )
    .pipe(m.do_hackernews_fetcher_udxf)
    .pipe(
        train_test_splits,
        unique_key="id",
        test_sizes=(0.9, 0.1),
        random_seed=0,
    )
)


# uncached run
(deferred_model, model_udaf, deferred_transform) = deferred_fit_transform_tfidf(
    train_expr,
)
model = deferred_model.execute()
transformed = test_expr.mutate(**{"transformed": deferred_transform.on_expr}).execute()


# cached run
storage = ParquetStorage(source=con)
(deferred_model, model_udaf, deferred_transform) = deferred_fit_transform_tfidf(
    train_expr, storage=storage
)
((cached_model,),) = deferred_model.execute().values
cached_transformed = test_expr.mutate(
    **{"transformed": deferred_transform.on_expr}
).execute()


assert transformed.equals(cached_transformed)
(x, y) = (pickle.loads(el) for el in (model, cached_model))
assert all(x.idf_ == y.idf_)
assert x.vocabulary_ == y.vocabulary_



import xorq as xo
from xorq import _
from xorq.caching import ParquetStorage


pg = xo.postgres.connect_examples()
con = xo.connect()

for table_name in pg.list_tables():
    if table_name.startswith(xo.config.options.cache.key_prefix):
        pg.drop_table(table_name)

cache = ParquetStorage(source=con)

t = (
    pg.table("batting")
    .mutate(row_number=xo.row_number().over(group_by=[_.playerID], order_by=[_.yearID]))
    .filter(_.row_number == 1)
    .cache(storage=cache)
)
print(f"{t.ls.get_key()} exists?: {t.ls.exists()}")
res = xo.execute(t)
print(res)
print(f"{t.ls.get_key()} exists?: {t.ls.exists()}")



import xorq as xo
from xorq import _
from xorq.caching import SourceStorage


con = xo.connect()
ddb = xo.duckdb.connect()
pg = xo.postgres.connect_env()

name = "batting"

right = (
    xo.examples.get_table_from_name(name, backend=ddb)
    .filter(_.yearID == 2014)
    .into_backend(con)
)
left = pg.table(name).filter(_.yearID == 2015).into_backend(con)

expr = left.join(
    right,
    "playerID",
).cache(SourceStorage(source=pg))

res = expr.execute()
print(res)



import argparse
import os
import sys
from pathlib import Path

from xorq.common.utils.import_utils import import_from_path
from xorq.ibis_yaml.compiler import BuildManager
from xorq.vendor.ibis import Expr


def build_command(script_path, expression, builds_dir="builds"):
    """
    Generate artifacts from an expression in a given Python script

    Parameters
    ----------
    script_path : Path to the Python script
    expression : The name of the expression to build
    builds_dir : Directory where artifacts will be generated

    Returns
    -------

    """

    if len(expression) != 1:
        print("Expected one, and only one expression", file=sys.stderr)
        sys.exit(1)

    (expression,) = expression

    if not os.path.exists(script_path):
        print(f"Error: Script not found at {script_path}", file=sys.stderr)
        sys.exit(1)

    print(f"Building {expression} from {script_path}")

    build_manager = BuildManager(builds_dir)

    vars_module = import_from_path(script_path)

    if not hasattr(vars_module, expression):
        print(f"Expression {expression} not found", file=sys.stderr)
        sys.exit(1)

    expr = getattr(vars_module, expression)

    if not isinstance(expr, Expr):
        print(
            f"The object {expression} must be an instance of {type(expr)}",
            file=sys.stderr,
        )
        sys.exit(1)

    expr_hash = build_manager.compile_expr(expr)
    print(
        f"Written '{expression}' to {build_manager.artifact_store.get_path(expr_hash)}"
    )


def run_command(expr_path, output_path=None, output_format="parquet"):
    """
    Execute an artifact

    Parameters
    ----------
    expr_path : str
        Path to the expr in the builds dir
    output_path : str
        Path to write output. Defaults to os.devnull
    output_format : str, optional
        Output format, either "csv", "json", or "parquet". Defaults to "parquet"

    Returns
    -------

    """
    if output_path is None:
        output_path = os.devnull

    try:
        expr_path = Path(expr_path)
        build_manager = BuildManager(expr_path.parent)
        expr = build_manager.load_expr(expr_path.stem)

        match output_format:
            case "csv":
                expr.to_csv(output_path)
            case "json":
                expr.to_json(output_path)
            case "parquet":
                expr.to_parquet(output_path)
            case _:
                raise ValueError(f"Unknown output_format: {output_format}")

    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)
        sys.exit(1)


def main():
    """Main entry point for the xorq CLI."""
    parser = argparse.ArgumentParser(description="xorq - build and run expressions")

    subparsers = parser.add_subparsers(dest="command", help="Command to execute")
    subparsers.required = True

    build_parser = subparsers.add_parser(
        "build", help="Generate artifacts from an expression"
    )
    build_parser.add_argument("script_path", help="Path to the Python script")
    build_parser.add_argument(
        "-e",
        "--expressions",
        nargs="?",
        help="Name of the expression variable in the Python script",
    )
    build_parser.add_argument(
        "--builds-dir", default="builds", help="Directory for all generated artifacts"
    )

    run_parser = subparsers.add_parser(
        "run", help="Run a build from a builds directory"
    )
    run_parser.add_argument("build_path", help="Path to the build script")
    run_parser.add_argument(
        "--output-path",
        default=None,
        help=f"Path to write output (default: {os.devnull})",
    )
    run_parser.add_argument(
        "--format",
        choices=["csv", "json", "parquet"],
        default="parquet",
        help="Output format (default: parquet)",
    )

    args = parser.parse_args()

    match args.command:
        case "build":
            expressions = [args.expressions] if args.expressions else []
            build_command(args.script_path, expressions, args.builds_dir)
        case "run":
            run_command(args.build_path, args.output_path, args.format)
        case _:
            raise ValueError(f"Unknown command: {args.command}")


if __name__ == "__main__":
    main()
